PROGRAM fcar_simulation

!
 INCLUDE 'CXML_INCLUDE.F90'     !necessary if using Compaq Visual Fortran
!                                 !includes BLAS libraries in compilation.
!  
!  USE define_types
!   USE f90_utilities
!   USE stat_utilities
!   USE ts_utilities
!  USE nlts_utilities
!  USE fcar_interfaces
   USE MSIMSL
!-----------------------------------------------------------------------
!  Declare variables
!-----------------------------------------------------------------------
!  Parameters.  IF larger sizes are needed, make changes here.  They
!  will carry throughout remainder of programs.
!  max_n: Largest allowable sample size ((n + extra) .LE. max_n).
!         Also serves as maximum number of bootstrap replications.
!  max_k: Largest allowable number of components in the vector 
!         time series - NOTE: IF max_k is > 9, change I1 in FORMAT
!         statements 10 and 20 at END of program to be IX where X is
!         the number of digits in max_k (k .LE max_k).
!  max_p: Largest allowable AR order in all models (p .LE max_p).
!  max_q: Largest allowable MA order in all models (q .LE max_q).
!  max_no_regimes:  Largest allowable number of regimes in TAR model
!                   (no_regimes .LE. max_no_regimes).
!  max_ns: Largest allowable number of sample sizes for study in 
!          simulation
!  max_corrs: Largest allowable number of error cross-correlations for
!             study in simulation.
!  bw_q: Number of subsections of data in bandwidth selection.
!  nh : Number of bandwidths to consider in optimal bandwidth selection.
!-----------------------------------------------------------------------
  INTEGER, PARAMETER :: max_n = 600, max_k = 5
  INTEGER, PARAMETER :: max_p = 5, max_q = 5
  INTEGER, PARAMETER :: max_no_regimes = 3
  INTEGER, PARAMETER :: bw_q = 4, nh = 50
  CHARACTER(30) :: outfile
!-----------------------------------------------------------------------
!  Simulation variables and arrays:
!-----------------------------------------------------------------------
  INTEGER :: i, j, l, no_ns, no_corrs, corr_no, n_no, no_bws, bw_no
  INTEGER :: no_reps, rep_no
  INTEGER, ALLOCATABLE, DIMENSION(:) :: ns
  REAL :: bw, alpha
  REAL, ALLOCATABLE, DIMENSION(:) :: corrs
  REAL, ALLOCATABLE, DIMENSION(:,:) :: bw_ave_mat
  REAL, ALLOCATABLE, DIMENSION(:,:,:,:) :: emp_rr
!-----------------------------------------------------------------------
!  Variables needed for all models:
!-----------------------------------------------------------------------
  INTEGER :: seed_option, model_option, extra, plusone
  INTEGER, DIMENSION(4) :: seed
!  Error mean, standard deviations, and cross-correlation.
  INTEGER :: sd_option, mean_option, intercept_option
  REAL, DIMENSION(max_k) :: sds, means
  REAL :: rho
!  Covariance, AR order, MA order, AR and MA coefficients
  INTEGER :: n, k, p, q
  REAL, DIMENSION(max_k, max_k):: cov
  REAL, DIMENSION(max_p, max_k, max_k) :: AR_coefs
  REAL, DIMENSION(max_q, max_k, max_k) :: MA_coefs
!-----------------------------------------------------------------------
!  Variables unique to vector bilinear:
!-----------------------------------------------------------------------
  INTEGER :: cp_option, cp_dim
  REAL, DIMENSION(max_k, ((max_p*max_k)**2)) :: cp_coefs
!-----------------------------------------------------------------------
!  Variables unique to vector threshold ARMA:
!-----------------------------------------------------------------------
  INTEGER :: no_regimes, thr_resp_ind, thr_lag
  INTEGER, DIMENSION(max_p) :: AR_lags
  INTEGER, DIMENSION(max_q) :: MA_lags
  REAL, DIMENSION(max_no_regimes) :: thresholds
  REAL, DIMENSION(max_no_regimes, max_k, max_k) :: TAR_covs
  REAL, DIMENSION(max_no_regimes, max_p, max_k, max_k) :: TAR_AR_coefs
  REAL, DIMENSION(max_no_regimes, max_q, max_k, max_k) :: TAR_MA_coefs
!-----------------------------------------------------------------------
!  Variables unique to vector smoothed threshold ARMA:
!-----------------------------------------------------------------------
  INTEGER :: sm_funct
  REAL :: sm_const, sm_param
!-----------------------------------------------------------------------
!  Variables unique to vector exponential AR:
!-----------------------------------------------------------------------
  INTEGER :: expar_delay
  REAL, DIMENSION(max_p, max_k, max_k) :: AR_cons
  REAL, DIMENSION(max_p, max_k*max_k, max_k) :: expar_coefs
!-----------------------------------------------------------------------
!  Vector time series realization matrix:
!-----------------------------------------------------------------------
  INTEGER :: t, dim, beta_dim
  REAL, DIMENSION(max_n, max_k) :: x
!-----------------------------------------------------------------------
!  AIC and BIC vectors:
!-----------------------------------------------------------------------
  INTEGER :: aic_option, iii
  INTEGER :: AR_fit_order, func_resp_ind, func_delay, max_lag, no_obs
  INTEGER :: no_cols, beg_sub, end_sub, max_order, bias_option
  INTEGER, DIMENSION(1) :: ind
  INTEGER, DIMENSION(max_p) :: orders
  REAL, DIMENSION(max_p) :: aic, bic
!-----------------------------------------------------------------------
!  Functional variable fitting:
!-----------------------------------------------------------------------
  INTEGER :: smoother, start, index_of_u
  INTEGER, DIMENSION(max_n) :: u_index
  REAL :: opt_bw, trim
  REAL, DIMENSION(max_n) :: func_var, u
  REAL, DIMENSION(2*max_n) :: z
  REAL, DIMENSION(max_n, max_k) :: y, y_s
  REAL, DIMENSION(max_n, max_k*max_p+1) :: xr, xr_s
  REAL, DIMENSION(max_n, max_p*max_k, max_k) :: f_hat, f_hat_s
!-----------------------------------------------------------------------
!  Bootstrap variables:
!-----------------------------------------------------------------------
  INTEGER :: no_bs_reps
  REAL :: rss0_det, rss1_det, T0, p_value
  REAL, DIMENSION(max_n) :: ts_star
  REAL, DIMENSION(max_k, max_k) :: rss0, rss1
  REAL, DIMENSION(max_n, max_k) :: y0_hat, y1_hat
  REAL, DIMENSION(max_n, max_k) :: y0_hat_s, y1_hat_s
  REAL, DIMENSION(max_k*max_p+1, max_k) :: beta_hat
!-----------------------------------------------------------------------
!  Bandwidth selection variables:
!-----------------------------------------------------------------------
  REAL :: range, step, band_width
  REAL, DIMENSION(nh+1) :: ams, bw_vec
!-----------------------------------------------------------------------
!  END declare variables.
!  Begin user input.
!-----------------------------------------------------------------------
!
  WRITE(*,*) "What is the name of the output file?"
  READ(*,*) outfile
!
  OPEN(100,file=outfile)
  WRITE(100,600)
  WRITE(100,610)
  WRITE(100,600)
!
  WRITE(*,*) "Enter a seed for the random number generator?"
  WRITE(*,*) "   (1 = Yes)"
  READ(*,*) seed_option
  IF(seed_option == 1) THEN
     seed = 0
     WRITE(*,*) "What is the seed?"
     READ(*,*) seed(1)
     WRITE(100,630) seed(1)
     CALL RANDOM_SEED(put = seed)
  ELSE
     WRITE(100,640)
  END IF
  WRITE(100,*)
!
  WRITE(*,*) "How many sample sizes?"
  READ(*,*) no_ns
  WRITE(100,740) no_ns
  WRITE(100,*)
  ALLOCATE(ns(no_ns))
  DO i = 1, no_ns
10   WRITE(*,540) i
     READ(*,*) ns(i)
     IF(ns(i) .GT. max_n) THEN
        WRITE(*,550) max_n
        GO TO 10
     END IF
  END DO
!
  WRITE(*,*) "How many extra (start-up) observations?"
  READ(*,*) extra
!
  WRITE(*,*) "What is the dimension of the time series?"
  READ(*,*) k
  WRITE(100,655) k
  WRITE(100,*)
!
  WRITE(*,*) "What model should be used to generate the data?"
  WRITE(*,*) " (0 => white noise,"
  WRITE(*,*) "  1 => ARMA,"
  WRITE(*,*) "  2 => bilinear,"
  WRITE(*,*) "  3 => threshold ARMA,"
  WRITE(*,*) "  4 => smoothed threshold AR,"
  WRITE(*,*) "  5 => exponential AR.)"
  READ(*,*) model_option
!
  IF(model_option .NE. 3) THEN
     WRITE(*,*) "How many error cross-correlations?"
     READ(*,*) no_corrs
     WRITE(100,745) no_corrs
     WRITE(100,*)
     ALLOCATE(corrs(no_corrs))
     DO i = 1, no_corrs
20      WRITE(*,560) i
        READ(*,*) corrs(i)
        IF(ABS(corrs(i)) .GT. 1.0) THEN
           WRITE(*,570)
           GO TO 20
        END IF
     END DO
     WRITE(*,*) "Should all standard deviations be 1? (1 => Yes)"
     READ(*,*) sd_option
     sds = 1.0
     IF(sd_option .NE. 1) THEN
        DO i = 1,k
           WRITE(*,520) i
           READ(*,*) sds(i)
        END DO
     END IF
  END IF
!
  IF(model_option == 0) THEN
     WRITE(100,620) k
     WRITE(*,*) "Should the mean be zero (1 => yes)?"
     READ(*,*) mean_option
     means = 0.0
     IF(mean_option .NE. 1) THEN
        DO i = 1,k
           WRITE(*,530) i
           READ(*,*) means(i)
        END DO
     END IF
  ELSE IF(model_option == 1) THEN
     AR_coefs = 0.0
     MA_coefs = 0.0
     CALL get_mvARMA_params(max_k,max_p,max_q,100,k,p,q,AR_coefs,MA_coefs)
  ELSE IF(model_option == 2) THEN
     CALL get_mvbilinear_params(max_k,max_p,max_q,100,k,p,q,cp_option,       &
          AR_coefs,MA_coefs,cp_dim,cp_coefs)
  ELSE IF(model_option == 3) THEN
     no_corrs = 1
     CALL get_mvTAR_params(max_k,max_p,max_q,max_no_regimes,100,k,p,q,       &
          no_regimes,TAR_AR_coefs,TAR_MA_coefs,AR_lags,MA_lags,thr_resp_ind, &
          thr_lag,thresholds,TAR_covs)
  ELSE IF(model_option == 4) THEN
     CALL get_mvSTAR_params(max_k,max_p,max_no_regimes,100,k,p,sm_funct,  &
          sm_const,sm_param,TAR_AR_coefs,AR_lags,thr_resp_ind,thr_lag) 
  ELSE IF(model_option == 5) THEN
     CALL get_mvEXPAR_params(max_k,max_p,100,k,p,AR_cons,AR_coefs, &
          expar_coefs,expar_delay)
  END IF
  WRITE(100,*)
!
  WRITE(*,*) "Use AIC or BIC to select AR fit order? (1 => Yes)"
  READ(*,*) aic_option
  IF(aic_option == 1) THEN
     WRITE(*,*) "Use AIC (0) or BIC (1)?"
     READ(*,*) bias_option
     WRITE(*,*) "What is the maximum AR fit order to be allowed?"
     READ(*,*) max_order
     IF(bias_option == 0) WRITE(100,768) 
     IF(bias_option == 1) WRITE(100,769) 
     WRITE(100,*)
  ELSE
     WRITE(*,*) "What is the AR fit order?"
     READ(*,*) AR_fit_order
     WRITE(100,770) AR_fit_order
  END IF
  WRITE(100,*)
!
  WRITE(*,*) "Include an intercept term when fitting the FCAR model? (1 => Yes)"
  READ(*,*) intercept_option
  IF(intercept_option .NE. 1) THEN
     plusone = 0
     WRITE(100,771)
  ELSE IF(intercept_option == 1) THEN
     plusone = 1
     WRITE(100,772)
  END IF
!
25 WRITE(*,*) "What percentage of FCAR estimates to trim for inference?"
  READ(*,*) trim
  IF(trim .LT. 0.0 .OR. trim .GT. 20.0) THEN
     WRITE(*,*) "ERROR: The trimming percentage must be between 0 and 20."
     GOTO 25
  END IF
!
  WRITE(*,*) "How many bootstrap replications?"
  READ(*,*) no_bs_reps
  WRITE(100,660) no_bs_reps
  WRITE(100,*)
!
  WRITE(*,*) "How many simulation replications?"
  READ(*,*) no_reps
  WRITE(100,670) no_reps
  WRITE(100,*)
!
  WRITE(*,*) "Enter nominal level of significance"
  READ(*,*) alpha
  WRITE(100,680) alpha
  WRITE(100,690)
  WRITE(100,600)
!-----------------------------------------------------------------------
!   End of user input.
!   Begin simulation:
!-----------------------------------------------------------------------
!   Correlation loop:
!-----------------------------------------------------------------------
  ALLOCATE(bw_ave_mat(MAX(p,q,2), k))
  ALLOCATE(emp_rr(MAX(p,q,2), k, no_ns, no_corrs))
  bw_ave_mat = 0.0
  emp_rr = 0.0
!
  x = 0.0
  DO corr_no = 1, no_corrs
     IF(model_option .NE. 3) rho = corrs(corr_no)
     WRITE(100,720) rho
     IF(model_option .NE. 3) THEN
        DO i = 1,k
           DO j = i,k
              IF(i == j) THEN
                 cov(i,j) = sds(i)*sds(i)
              ELSE
                 cov(i,j) = sds(i)*sds(j)*rho
                 cov(j,i) = cov(i,j)
              END IF
           END DO
        END DO
        WRITE(100,*)
        WRITE(100,730)
        DO i = 1, k
           WRITE(100,*) (cov(i,j), j = 1,k)
        END DO
     END IF
     WRITE(100,600)
!-----------------------------------------------------------------------
!   Sample size loop:
!-----------------------------------------------------------------------
     DO n_no = 1, no_ns
        n = ns(n_no)
        nn = n + extra
        WRITE(100,775) n
        WRITE(100,815)
        DO rep_no = 1, no_reps
!-----------------------------------------------------------------------
!   Generate realization:
!-----------------------------------------------------------------------
           IF(model_option == 0) THEN
              CALL gen_normal(max_n,max_k,n,k,means,cov,1,x)
           ELSE IF(model_option == 1) THEN
              CALL gen_mvARMA(max_n,max_k,max_p,max_q,n,k,p,q,            &
                   AR_coefs,MA_coefs,cov,1,extra,x)
           ELSE IF(model_option == 2) THEN
              CALL gen_mvbilinear(max_n,max_k,max_p,max_q,n,k,p,q,        &
                   AR_coefs,MA_coefs,cp_option,cp_dim,cp_coefs,1,cov,     &
                   extra,x)
           ELSE IF(model_option == 3) THEN
              CALL gen_mvTAR(max_n,max_k,max_p,max_q,max_no_regimes,n,    &
                   k,p,q,no_regimes,TAR_AR_coefs,TAR_MA_coefs,AR_lags,    &
                   MA_lags,thr_resp_ind,thr_lag,thresholds,TAR_covs,1,    &
                   extra,x)
           ELSE IF(model_option == 4) THEN
              CALL gen_mvSTAR(max_n,max_k,max_p,max_no_regimes,n,k,p,     &
                   sm_funct,sm_const,sm_param,TAR_AR_coefs,AR_lags,    &
                   thr_resp_ind,thr_lag,cov,1,extra,x)  
           ELSE IF(model_option == 5) THEN
              CALL gen_mvEXPAR(max_n,max_k,max_p,n,k,p,AR_cons,AR_coefs,  &
                   expar_coefs,expar_delay,cov,1,extra,x)
           END IF
!-----------------------------------------------------------------------
!   Determine AR fit order if requested (aic_option == 1):
!-----------------------------------------------------------------------
           IF(aic_option == 1) THEN
              orders = 0
              aic = 0.0
              bic = 0.0
              CALL mvaic(max_n,n,k,x,max_order,orders,aic,bic)
              IF(bias_option == 1) THEN
                 ind = MINLOC(bic(1:max_order))
                 AR_fit_order = orders(ind(1))
              ELSE
                 ind = MINLOC(aic(1:max_order))
                 AR_fit_order = orders(ind(1))
              END IF
           END IF
!-----------------------------------------------------------------------
!   Functional variable loop (response index and delay):
!-----------------------------------------------------------------------
           IF(intercept_option == 1) xr(1:n,1) = 1.0
           no_cols = k*AR_fit_order + plusone
           DO func_resp_ind = 1, k
              band_width = 0.0
              DO func_delay = 1, MAX(p,q,2)
                 max_lag = MAX(AR_fit_order, func_delay)
                 no_obs = n - max_lag
                 y(1:no_obs,1:k) = x((max_lag+1):n,1:k)                 
                 DO i = 1,AR_fit_order
                    xr(1:no_obs,(k*(i-1)+(1+plusone)):(i*k+plusone)) = &
					    x((max_lag-i+1):(n-i),1:k)
                 END DO
!
                 beg_sub = 1 + max_lag - func_delay
                 end_sub = no_obs + max_lag - func_delay
                 func_var(1:no_obs) = x(beg_sub:end_sub,func_resp_ind)
!-----------------------------------------------------------------------
!   Compute fitted values from linear regression:
!-----------------------------------------------------------------------
                 beta_dim = max_k*max_p+1
                 CALL ls_regression(max_n,no_obs,k,no_cols,xr,y,beta_dim,beta_hat)
                 y0_hat(1:no_obs,1:k) = MATMUL(xr(1:no_obs,1:no_cols),beta_hat(1:no_cols,1:k))
!-----------------------------------------------------------------------
!   Compute bandwidth, fit FCAR model to data, and compute fitted
!   values from FCAR model.
!-----------------------------------------------------------------------
                 CALL fcar_bandwidth(max_n,no_obs,k,no_cols,max_p*max_k,y,xr, &
                      func_var,func_var,3,nh,bw_q,ams,bw_vec)
                 ind = MINLOC(ams)
                 opt_bw = bw_vec(ind(1))
                 CALL fit_fcar(max_n,no_obs,k,no_cols,y,xr,opt_bw,func_var,   &
                      func_var,3,0,no_obs,max_p*max_k,f_hat)
                 y1_hat(1:no_obs,1:k) = 0.0
                 DO j = 1, no_obs
                 DO l = 1, no_cols
                    y1_hat(j,1:k) = y1_hat(j,1:k) + xr(j,l)*f_hat(j,l,1:k)
                 END DO
                 END DO
!-----------------------------------------------------------------------
!   Sort values according to the ordered values of the functional
!   variable.
!-----------------------------------------------------------------------
                 DO i = 1, no_obs
                    u_index(i) = i
                 END DO
                 z(1:no_obs) = u(1:no_obs)
                 CALL index_sort(no_obs,u,u_index,z)
                 DO i = 1, no_obs
                    index_of_u = u_index(i)
                    y_s(i,1:k) = y(index_of_u,1:k)
                    y0_hat_s(i,1:k) = y0_hat(index_of_u,1:k)
                    y1_hat_s(i,1:k) = y1_hat(index_of_u,1:k)
                    xr_s(i,1:no_cols) = xr(index_of_u,1:no_cols)
                 END DO
!
                 y0_hat_s(1:no_obs,1:k) = y_s(1:no_obs,1:k) - y0_hat_s(1:no_obs,1:k)
                 y1_hat_s(1:no_obs,1:k) = y_s(1:no_obs,1:k) - y1_hat_s(1:no_obs,1:k)
                 T0 = 0.0
                 CALL fcar_test(max_n,max_k,no_obs,k,trim,y0_hat_s,y1_hat_s, &
                      rss0,rss1,rss0_det,rss1_det,T0)
!
!  The unsorted y1_hats are replaced with the residuals from the FCAR model
!  for use in bootstrap_test.
!
                 y1_hat(1:no_obs,1:k) = y(1:no_obs,1:k) - y1_hat(1:no_obs,1:k)

                 CALL bootstrap_test(max_n,max_k,no_obs,k,no_cols,trim,y,xr, &
                      func_var,func_var,y0_hat,y1_hat,opt_bw,no_bs_reps,     &
                      rss0_det,rss1_det,T0,ts_star,p_value)
                 WRITE(100,820) rep_no, func_delay, func_resp_ind,            &
                      AR_fit_order, n, rho, T0, p_value
                 IF(p_value .LT. alpha) THEN
                    emp_rr(func_delay,func_resp_ind,n_no,corr_no) =  1.0 + &
                         emp_rr(func_delay, func_resp_ind, n_no, corr_no)
                 END IF
              END DO            ! End func_delay loop
           END DO               ! End func_resp_ind loop
        END DO                  ! End no_rep loop
      END DO                    ! End n_no loop
  END DO                        ! End corr_no loop
  WRITE(100,600)
!
  emp_rr = emp_rr/DBLE(no_reps)
!
  WRITE(100,600)
  WRITE(100,830)
  WRITE(100,840)
  WRITE(100,600)
  DO corr_no = 1, no_corrs
     IF(model_option .NE. 3) rho = corrs(corr_no)
     DO n_no = 1, no_ns
        n = ns(n_no)
        DO func_resp_ind = 1, k
           DO func_delay = 1, MAX(p,q,2)
              WRITE(100,850) func_delay, func_resp_ind, n, rho,   &
                   emp_rr(func_delay, func_resp_ind, n_no, corr_no)
           END DO
        END DO
     END DO
  END DO
  WRITE(100,600)
  CLOSE(100)
  DEALLOCATE(ns, bw_ave_mat, emp_rr)
  IF(model_option .NE. 3) DEALLOCATE(corrs)
!-----------------------------------------------------------------------
!   Format statements for variable input (520 - 570)
!   and for results output (600 - 740)
!-----------------------------------------------------------------------
520 FORMAT("Enter standard deviation ",I1,".")
530 FORMAT("Enter mean ",I1,".")
540 FORMAT("Enter sample size ",I1,".")
550 FORMAT("All sample sizes must be less than ",I6,".")
560 FORMAT("Enter error cross-correlation ",I1,".")
570 FORMAT("All error cross-correlations must be between -1 and +1.")
!-----------------------------------------------------------------------
600 FORMAT("******************************************************************")
610 FORMAT(5X,"Generate realization from vector time series")
620 FORMAT("Vector random normal process of dimension ",I1,".")
630 FORMAT("User-specified seed: ",I8,".")
640 FORMAT("Seed not specified.")
650 FORMAT("Number of extra (start-up) observations generated: ",I5)
655 FORMAT("Time series dimension: ",I1,".")
660 FORMAT("Number of bootstrap replications: ", I4,".")
670 FORMAT("Number of simulation replications: ", I5,".")
680 FORMAT("Nominal level of significance: ",F6.4,".")
690 FORMAT("End of user input.")
700 FORMAT("Bootstrap centered residuals.")
710 FORMAT("Bootstrap centered residuals and compute bootstrap sample.")
712 FORMAT("Test statistic = RSS_0/RSS_1 - 1.")
715 FORMAT("Test statistic = likelihood ratio.")
720 FORMAT("Error cross-correlation: ",F8.6)
730 FORMAT("Covariance matrix:")
740 FORMAT("Number of sample sizes: ",I2,".")
745 FORMAT("Number of error cross-correlations: ",I1,".")
750 FORMAT("BIC selected AR fit order: ",I1,".")
760 FORMAT("AIC selected AR fit order: ",I1,".")
768 FORMAT("AIC used to select AR fit order.")
769 FORMAT("BIC used to select AR fit order.")
770 FORMAT("User-specified AR fit order: ",I1,".")
771 FORMAT("Intercept was not included when fitting FCAR model.")
772 FORMAT("Intercept was included when fitting FCAR model.")
775 FORMAT("Sample size: ",I6,".")
780 FORMAT("Functional variable: X_",I1,"(t - ",I1,").")
800 FORMAT("Resp. Ind., Delay, Bandwidth")
810 FORMAT(5X,I2,6X,I2,2X,F10.6)
815 FORMAT("Rep. no., Delay, Resp. ind., AR ord., n, rho, test stat., p-value")
820 FORMAT(I5,2X,I1,3X,I1,3X,I1,3X,I6,2X,F6.4,2X,F12.6,2X,F12.6)
830 FORMAT("Empirical Rejection Rates")
840 FORMAT("Delay, Resp. ind., n,   rho,     Emp. rr.")
850 FORMAT(2X,I1,9X,I1,4X,I6,1X,F6.4,1X,F8.6)
!-----------------------------------------------------------------------
END PROGRAM fcar_simulation


SUBROUTINE fcar_bandwidth(max_n,n,k,p,f_dim,y,x,u,z,smoother,nh,bw_q,ams,bw_vec)
!***********************************************************************
!  FORTRAN 90 subroutine to compute the band width for FCAR modelling.
!  Notes: Uses BLAS function ddot and subroutine dgemv.  Both are
!         available for free download from http://www.netlib.org/blas/
!***********************************************************************
!  USE define_types
!
  IMPLICIT NONE
  INTEGER, INTENT(in) :: max_n, n, k, p, f_dim, smoother, bw_q, nh
  REAL, INTENT(in), DIMENSION(max_n) :: u, z
  REAL, INTENT(in), DIMENSION(max_n, k) :: y
  REAL, INTENT(in), DIMENSION(max_n, p) :: x
  REAL, INTENT(inout), DIMENSION(nh+1) :: bw_vec, ams
!
  INTEGER :: m, n1, i, j, t, jj
  REAL :: dn, h, sse, maxu, minu, range_of_u, a, step
  REAL, DIMENSION(max_n,f_dim,k) :: f_hat
  REAL, DIMENSION(k) :: err, y_hat
!
  dn = DBLE(n)
  m = INT(0.1*dn)
  ams = 0.0
  maxu = MAXVAL(u(1:n))
  minu = MINVAL(u(1:n))
  range_of_u = maxu - minu
  a = 1.0/(DBLE(n)**0.20)
  step = range_of_u*a*1.5/(DBLE(nh))
  bw_vec(1) = 0.5*range_of_u*a
  DO i = 2, nh+1
     bw_vec(i) = bw_vec(i-1) + step
  END DO
!
  DO i = 1, bw_q
     n1 = n - i*m
     DO j = 1, nh+1
        h = bw_vec(j) * (dn/DBLE(n1))**0.2
        CALL fit_fcar(max_n,n1,k,p,y,x,h,u,z,smoother,n1,m,f_dim,f_hat)
        DO t = n1+1, n1+m
           y_hat = MATMUL(x(t,1:p),f_hat(t,1:p,1:k))
           err(1:k) = y(t,1:k) - y_hat(1:k)
           DO jj = 1, k
              ams(j) = ams(j) + err(jj)*err(jj)
           END DO
        END DO
     END DO
  END DO
!
RETURN
END SUBROUTINE fcar_bandwidth






MODULE ts_interfaces
!***********************************************************************
!  FORTRAN 90 module containing interfaces for traditional time series
!  analysis programs.
!***********************************************************************
  INTERFACE
     SUBROUTINE get_mvARMA_params(max_k,max_p,max_q,file_label,k,p,q,AR_coefs,MA_coefs)
!       USE define_types
       IMPLICIT NONE
       INTEGER, INTENT(in) :: max_k, max_p, max_q, file_label, k
       INTEGER, INTENT(out) :: p, q
       REAL, INTENT(inout), DIMENSION(max_p, max_k, max_k) :: AR_coefs
       REAL, INTENT(inout), DIMENSION(max_q, max_k, max_k) :: MA_coefs
     END SUBROUTINE get_mvARMA_params
  END INTERFACE
!
  INTERFACE
     SUBROUTINE gen_mvARMA(max_n,max_k,max_p,max_q,n,k,p,q,AR_coefs,MA_coefs,   &
          cov,cov_fac_option,extra,x)
!       USE define_types
       USE stat_utilities
       IMPLICIT NONE
       INTEGER, INTENT(in) :: max_n, max_k, max_p, max_q
       INTEGER, INTENT(in) :: n, k, p, q, cov_fac_option, extra
       REAL, INTENT(in), DIMENSION(max_p, max_k, max_k) :: AR_coefs
       REAL, INTENT(in), DIMENSION(max_q, max_k, max_k) :: MA_coefs
       REAL, INTENT(in), DIMENSION(max_k, k) :: cov
       REAL, INTENT(inout), DIMENSION(max_n, k) :: x
     END SUBROUTINE gen_mvARMA
  END INTERFACE
!
  INTERFACE
     SUBROUTINE mvaic(max_n,n,k,x,max_order,orders,aic,bic)
!       USE define_types
       USE stat_utilities
       IMPLICIT NONE
       INTEGER, INTENT(in) :: max_n, n, k, max_order
       REAL, INTENT(in), DIMENSION(max_n, k) :: x
       INTEGER, INTENT(out), DIMENSION(max_order) :: orders
       REAL, INTENT(out), DIMENSION(max_order) :: aic, bic
       EXTERNAL dgetrf
     END SUBROUTINE mvaic
  END INTERFACE
!
END MODULE ts_interfaces


MODULE ts_utilities
!***********************************************************************
!   FORTRAN 90 module containing commonly used time series
!   statistical computations.
!
!   Created: 03/12/03 JLH
!
!   Contents}:
!        Subroutines for generating vector ARMA realizations:
!            get_mvARMA_params, gen_mvARMA
!        Subroutines for computing optimal AR order:
!             mvaic
!   NOTES: mvaic depends on LAPACK subroutines for computing matrix
!          inverse and LU factorizations for determinant.
!***********************************************************************
!  USE define_types
  USE stat_utilities
!
  IMPLICIT NONE
!***********************************************************************
!  Generate mvARMA:
!***********************************************************************
!  Get coefficients:
!***********************************************************************
CONTAINS
  SUBROUTINE get_mvARMA_params(max_k,max_p,max_q,file_label,k,p,q,AR_coefs,MA_coefs)
!    USE define_types
    IMPLICIT NONE
!
    INTEGER, INTENT(in) :: max_k, max_p, max_q, file_label, k
    INTEGER, INTENT(out) :: p, q
    REAL, INTENT(inout), DIMENSION(max_p, max_k, max_k) :: AR_coefs
    REAL, INTENT(inout), DIMENSION(max_q, max_k, max_k) :: MA_coefs
!
    INTEGER :: lag, dim1, dim2
!
    WRITE(file_label,400) k
    WRITE(*,*) "Enter the AR order (p) and MA order (q)."
    WRITE(*,*) "p = ?"
    READ(*,*) p
    WRITE(*,*) "q = ?"
    READ(*,*) q
    WRITE(file_label,500) p, q
!
    IF(p > 0) THEN
       WRITE(*,*) "Enter the AR coefficients, row by row."
       WRITE(file_label,510) 
       DO lag = 1, p
          WRITE(*,520) lag
          WRITE(file_label,520) lag
          DO dim1 = 1, k
             READ(*,*) (AR_coefs(lag,dim1,dim2), dim2 = 1, k)
             WRITE(file_label,*) (AR_coefs(lag,dim1,dim2), dim2 = 1, k)
          END DO
       END DO
    END IF
!
    IF(q > 0) THEN
       WRITE(*,*) "Enter the MA coefficients, row by row."
       WRITE(file_label,530) 
       DO lag = 1, q
          WRITE(*,520) lag
          WRITE(file_label,520) lag
          DO dim1 = 1, k
             READ(*,*) (MA_coefs(lag,dim1,dim2), dim2 = 1, k)
             WRITE(file_label,*) (MA_coefs(lag,dim1,dim2), dim2 = 1, k)
          END DO
       END DO
    END IF
!
400 FORMAT("Vector autregressive moving average process of dimension ",I1,".")
500 FORMAT("AR order (p) = ",I2,": MA order (q) = ",I2,".")
510 FORMAT("The AR coefficients (row by row) are:")
520 FORMAT("Lag ",I2,":")
530 FORMAT("The MA coefficients (row by row) are:")
!
    RETURN
  END SUBROUTINE get_mvARMA_params
!***********************************************************************
!  Generate realization:
!***********************************************************************
  SUBROUTINE gen_mvARMA(max_n,max_k,max_p,max_q,n,k,p,q,AR_coefs,MA_coefs,    &
       cov,cov_fac_option,extra,x)
!    USE define_types
    USE stat_utilities
!
    IMPLICIT NONE
!
    INTEGER, INTENT(in) :: max_n, max_k, max_p, max_q
    INTEGER, INTENT(in) :: n, k, p, q, cov_fac_option, extra
    REAL, INTENT(in), DIMENSION(max_p, max_k, max_k) :: AR_coefs
    REAL, INTENT(in), DIMENSION(max_q, max_k, max_k) :: MA_coefs
    REAL, INTENT(in), DIMENSION(max_k, k) :: cov
    REAL, INTENT(inout), DIMENSION (max_n, k) :: x
!
    INTEGER :: t, j, info
    REAL, DIMENSION(:), ALLOCATABLE :: mean_vec
    REAL, DIMENSION(:,:), ALLOCATABLE :: eps, coefs, work_space
!
    ALLOCATE(eps(max_n,k))
    ALLOCATE(mean_vec(k))
    mean_vec = 0.0
    CALL gen_normal(max_n,max_k,(n+extra),k,mean_vec,cov,cov_fac_option,eps)
    DEALLOCATE(mean_vec)
!
    x = eps
!
    ALLOCATE(work_space(1,k))
    work_space = 0.0
    DO t = p+1,(n+extra)
       DO j = 1, MIN(t-1,p)
          work_space(1,1:k) = MATMUL(x(t-j,1:k),AR_coefs(j,1:k,1:k))
          x(t,1:k) = x(t,1:k) + work_space(1,1:k)
       END DO
       DO j = 1, MIN(t-1,q)
          work_space(1,1:k) = MATMUL(eps(t-j,1:k),MA_coefs(j,1:k,1:k))
          x(t,1:k) = x(t,1:k) - work_space(1,1:k)
       END DO
    END DO
!
    x(1:n,1:k) = x((extra+1):(extra+n),1:k)
    DEALLOCATE(work_space, eps)
!
    RETURN
  END SUBROUTINE gen_mvARMA
!***********************************************************************
!   FORTRAN 90 subroutine for computing AIC and BIC.
!   NOTES: 1.  Uses LAPACK subroutine spotrf (or dpotrf) to compute the 
!          Cholesky decomposition of the covariance matrix.  If LAPACK
!          not available, the Cholesky factorization subroutine in
!          f90_utilities can be used by commenting out the lines
!          EXTERNAL spotrf and IF(cov_fac_option .NE. 0) call spotrf(.)
!          and uncommenting the four lines preceding it.
!
!          2.  Uses LAPACK subroutines dgetrf and dgetri to compute
!          (X^t X)^{-1} in finding least squares estimates for regression
!          coefficients.
!***********************************************************************
  SUBROUTINE mvaic(max_n,n,k,x,max_order,orders,aic,bic)
!    USE define_types
    USE stat_utilities
!
    IMPLICIT NONE
!
    INTEGER, INTENT(in) :: max_n, n, k, max_order
    REAL, INTENT(in), DIMENSION(max_n, k) :: x
    INTEGER, INTENT(out), DIMENSION(max_order) :: orders
    REAL, INTENT(out), DIMENSION(max_order) :: aic, bic
!
    INTEGER :: order, no_obs, col_x, last_col_x
    INTEGER :: t, dim, lag, index, info, ii, jj
    INTEGER, DIMENSION(:), ALLOCATABLE :: pivot
    REAL :: dk, dno_obs, det_cov, d, k2o
    REAL, DIMENSION(:,:), ALLOCATABLE :: work_space, beta_hat
    REAL, DIMENSION(:,:), ALLOCATABLE :: fitted, residuals, cov_hat
!
    EXTERNAL dgetrf
!
    dk = DBLE(k)
    aic = 0.0
    bic = 0.0
    orders = 0
!
    ALLOCATE(cov_hat(k,k))
    DO order = 1, max_order
       no_obs = n - order
       col_x = k*order
       last_col_x = k + col_x
       orders(order) = order
!
       ALLOCATE(work_space(no_obs,last_col_x))
       DO t = 1, no_obs
          work_space(t,1:k) = x(t+order,1:k)
          DO dim = 1, k
             DO lag = 1, order
                index = k*lag + dim
                work_space(t,index) = x(t+lag-1,dim)
             END DO
          END DO
       END DO
!
       ALLOCATE(beta_hat(col_x,k))
       ALLOCATE(fitted(no_obs,k))
       ALLOCATE(residuals(no_obs,k))
!
       CALL ls_regression(max_n,n,k,col_x,work_space(1:no_obs,(k+1):last_col_x),  &
	        work_space(1:no_obs,1:k),col_x,beta_hat)
!
       fitted = MATMUL(work_space(1:no_obs,(k+1):last_col_x),beta_hat)
       residuals = work_space(1:no_obs,1:k) - fitted(1:no_obs,1:k)
       cov_hat = MATMUL(TRANSPOSE(residuals),residuals)
!
       ALLOCATE(pivot(k))
       CALL dgetrf(k,k,cov_hat,k,pivot,info)       
       DEALLOCATE(pivot)
!
       det_cov = 1.0
       DO dim = 1,k
          det_cov = det_cov*ABS(cov_hat(dim,dim))
       END DO
!
       DEALLOCATE(work_space, beta_hat, fitted, residuals)
!
       dno_obs = DBLE(no_obs)
       d = dno_obs * LOG(det_cov) 
       k2o = dk * dk * DBLE(order)
!
       aic(order) = d + 2.0*k2o
       bic(order) = d + LOG(dno_obs)*k2o
    END DO
    DEALLOCATE(cov_hat)
!
    RETURN
  END SUBROUTINE mvaic
!
END MODULE ts_utilities



MODULE stat_utilities
!***********************************************************************
!   FORTRAN 90 module containing commonly used statistical computations.
!
!   Created: 03/01/03 JLH
!
!   LAPACK subroutines xpotrf, xgetrf, xgetri are used in some 
!   subroutines.  These subroutines are available for free download
!   from http://www.netlib.org/lapack/.
!
!   BLAS subroutines xgemv and xgemm are used for matrix - vector and
!   matrix-matrix multiplication.  These are available for free 
!   download from http://www.netlib.org/blas/ .
!
!   In the names of the subroutines above, the x should be replaced
!   with a "s" for single precision or a "d" for double precision.
!
!   Contents:
!        Suboutines for generating random numbers:
!            gen_uniform, gen_normal
!        Smoothing functions:
!             logistic_sm, exponential_sm, quad_log_sm, quartic_sm,
!             Epanechnikov, cubic_sm
!        Subroutines for linear regression:
!             ls_regr_r, ls_regr_d, ls_regr_rmv, ls_regr_dmv
!***********************************************************************
!  USE define_types
! USE f90_utilities
!
  IMPLICIT NONE
!***********************************************************************
!  Uniform(a,b) deviates:
!***********************************************************************
  INTERFACE gen_uniform
     MODULE PROCEDURE gen_uniform_r   ! , gen_uniform_d
  END INTERFACE ! rlc gen_uniform
!***********************************************************************
!  Normal(mean, variance) deviates:
!***********************************************************************
  INTERFACE gen_normal
     MODULE PROCEDURE gen_normal_r, gen_normal_d, gen_mvnormal_r   
         ! gen_mvnormal_d
  END INTERFACE ! rlc gen_normal
!***********************************************************************
!  Smoothing functions:
!***********************************************************************
!     Logistic:
!***********************************************************************
  INTERFACE logistic_sm
     MODULE PROCEDURE logistic_sm_r  ! , logistic_sm_d
  END INTERFACE ! rlc logistic_sm
!***********************************************************************
!     Exponential:
!***********************************************************************
  INTERFACE exponential_sm
     MODULE PROCEDURE exponential_sm_r  ! , exponential_sm_d
  END INTERFACE ! rlc exponential_sm
!***********************************************************************
!     Quadratic logarithm:
!***********************************************************************
  INTERFACE quad_log_sm
     MODULE PROCEDURE quad_log_sm_r  ! , quad_log_sm_d
  END INTERFACE ! rlc quad_log_sm
!***********************************************************************
!     Quartic:
!***********************************************************************
  INTERFACE quartic_sm
     MODULE PROCEDURE quartic_sm_r ! , quartic_sm_d
  END INTERFACE ! rlc quartic_sm
!***********************************************************************
!     Epanechnikov:
!***********************************************************************
  INTERFACE Epanechnikov
     MODULE PROCEDURE Epanechnikov_r  !, Epanechnikov_d
  END INTERFACE ! rlc Epanechnikov
!***********************************************************************
!      Cubic:
!***********************************************************************
  INTERFACE cubic_sm
     MODULE PROCEDURE cubic_sm_r   ! , cubic_sm_d
  END INTERFACE ! rlc cubic_sm
!***********************************************************************
!   Linear regression:
!***********************************************************************
  INTERFACE ls_regression
     MODULE PROCEDURE ls_regr_r, ls_regr_d, ls_regr_rmv  ! , ls_regr_dmv
  END INTERFACE ! rlc ls_regression
!
CONTAINS
!
  SUBROUTINE gen_uniform_r(a,b,n,x)
!***********************************************************************
!  FORTRAN 90 subroutine to generate n uniform(a,b) random variates.
!***********************************************************************
!    USE define_types
!
    IMPLICIT NONE
!
    INTEGER, INTENT(in) :: n
    REAL, INTENT(in) :: a, b
    REAL, INTENT(out), DIMENSION(n) :: x
!
    INTEGER :: i
    REAL :: c
!
    DO i = 1,n
       CALL RANDOM_NUMBER(c)
       x(i) = (b - a)*c + a
    END DO
!
    RETURN
  END SUBROUTINE gen_uniform_r
!
  SUBROUTINE gen_uniform_d(a,b,n,x)
!***********************************************************************
!  FORTRAN 90 subroutine to generate n uniform(a,b) random variates.
!***********************************************************************
!    USE define_types
!
    IMPLICIT NONE
!
    INTEGER, INTENT(in) :: n
    REAL, INTENT(in) :: a, b
    REAL, INTENT(out), DIMENSION(n) :: x
!
    INTEGER :: i
    REAL :: c
!
    DO i = 1,n
       CALL RANDOM_NUMBER(c)
       x(i) = (b - a)*c + a
    END DO
!
    RETURN
  END SUBROUTINE gen_uniform_d
!
  SUBROUTINE gen_normal_r(mean,variance,n,x)
!***********************************************************************
!  FORTRAN 90 subroutine to generate n N(mean,variance)
!  random variates.
!***********************************************************************
!    USE define_types
!
    IMPLICIT NONE
!
    INTEGER, INTENT(in) :: n
    REAL, INTENT(in) :: mean, variance
    REAL, INTENT(out), DIMENSION(n) :: x
!
    INTEGER :: nn, nnb2, i, i1, i2
    REAL :: s, y
    REAL, DIMENSION(2) :: r
!  
    nn = n - 1
    IF(2*(n/2) .EQ. n) nn = n
    IF(n .EQ. nn) go to 20
!
5   CALL gen_uniform(0.0,1.0,2,r)
    !
    r = 2.0*(r - 0.50)
    s = DOT_PRODUCT(r,r)
    !
    IF(s .GT. 1.0) go to 5
    y = SQRT(-2.0*(LOG(s)/s))
    !
    x(n) = r(1)*y
    IF(n .EQ. 1) go to 99
    !
20  CALL gen_uniform(0.0,1.0,nn,x)
    nnb2 = nn/2
    !
    DO i = 1, nnb2
       i2 = 2*i
       i1 = i2 - 1
12     r(1) = 2.0 * (x(i1) - 0.50)
       r(2) = 2.0 * (x(i2) - 0.50)
       s = DOT_PRODUCT(r,r)
       IF(s .LE. 1.0) go to 15
       CALL gen_uniform(0.0,1.0,2,x(i1:i2))
       go to 12
15     y = SQRT(-2.0 * (LOG(s)/s))
       x(i1) = y*r(1)
       x(i2) = y*r(2)
    END DO
    x = SQRT(variance)*x + mean
!
99  CONTINUE
    RETURN
!
  END SUBROUTINE gen_normal_r
!
  SUBROUTINE gen_normal_d(mean,variance,n,x)
!***********************************************************************
!  FORTRAN 90 subroutine to generate n N(mean,variance)
!  random variates.
!***********************************************************************
!   USE define_types
!
    IMPLICIT NONE
!
    INTEGER, INTENT(in) :: n
    REAL, INTENT(in) :: mean, variance
    REAL, INTENT(out), DIMENSION(n) :: x
!
    INTEGER :: nn, nnb2, i, i1, i2
    REAL :: s, y
    REAL, DIMENSION(2) :: r
!  
    nn = n - 1
    IF(2*(n/2) .EQ. n) nn = n
    IF(n .EQ. nn) go to 20
!
5   CALL gen_uniform(0.0,1.0,2,r)
!
    r = 2.0*(r - 0.50)
    s = DOT_PRODUCT(r,r)
!
    IF(s .GT. 1.0) go to 5
    y = SQRT(-2.0*(LOG(s)/s))
    !
    x(n) = r(1)*y
    IF(n .EQ. 1) go to 99
!
20  CALL gen_uniform(0.0,1.0,nn,x)
    nnb2 = nn/2
!
    DO i = 1, nnb2
       i2 = 2*i
       i1 = i2 - 1
12     r(1) = 2.0 * (x(i1) - 0.50)
       r(2) = 2.0 * (x(i2) - 0.50)
       s = DOT_PRODUCT(r,r)
       IF(s .LE. 1.0) go to 15
       CALL gen_uniform(0.0,1.0,2,x(i1:i2))
       go to 12
15     y = SQRT(-2.0 * (LOG(s)/s))
       x(i1) = y*r(1)
       x(i2) = y*r(2)
    END DO
    x = SQRT(variance)*x + mean
!
99  CONTINUE
    RETURN
!
  END SUBROUTINE gen_normal_d
!**********************************************************************
!  Multivariate normal:
!***********************************************************************
!  FORTRAN 90 subroutine to generate random normal observations having
!  mean vector mu and covariance matrix cov.  If cov.fac.opt = 0, then
!  the matrix cov contains the Cholesky decomposition of the covariance
!  matrix.  Otherwise, the Cholesky decomposition will be found.
!
!  Input: mean: a k-vector containing the means
!         cov: a k x k matrix containing the covariance matrix if
!              cov_fac_option != 0, and containing the Cholesky
!              decomposition of the covariance matrix otherwise.
!         cov_fac_option: an integer indicating whether or not the
!               matrix cov is the Cholesky decomposition of the 
!               covariance matrix (if 0 then no; otherwise, yes)
!
!  Output: x: an n x k matrix of multivariate normal deviates.
!
!  Algorigthm: Generate k iid E ~ N(0,1)
!              cov = Cholesky decomposition of covariance matrix
!              x( ,1:k) = trans(cov)E + mean ~ MVN(mean,covariance)
!
!  NOTES: uses LAPACK subroutine spotrf (or dpotrf) to compute the 
!         Cholesky decomposition of the covariance matrix.  See
!         http://www.netlib/org/lapack/.
!***********************************************************************
  SUBROUTINE gen_mvnormal_r(max_n,max_k,n,k,mean,cov,cov_fac_option,x)
!    USE define_types
!
    IMPLICIT NONE
!
    INTEGER, INTENT(in) :: max_n, max_k, n, k, cov_fac_option
    REAL, INTENT(in), DIMENSION(k) :: mean
    REAL, INTENT(in), DIMENSION(max_k, k) :: cov
    REAL, INTENT(out), DIMENSION(max_n, k) :: x
!
    INTEGER :: info, i, j
    REAL, DIMENSION(:), ALLOCATABLE :: temp, temp_out
    REAL, DIMENSION(:,:), ALLOCATABLE :: cov_fac
!
    EXTERNAL spotrf
!
    ALLOCATE(cov_fac(k,k))
    cov_fac = cov
    DO i = 1, k
    DO j = 1, i-1
       cov_fac(i,j) = 0.0
    END DO
    END DO
!
    IF(cov_fac_option .NE. 0) CALL spotrf("U",k,cov_fac,k,info)
!
    ALLOCATE(temp(k))
    ALLOCATE(temp_out(k))
    DO i = 1,n
       CALL gen_normal(0.0,1.0,k,temp)
       temp = MATMUL(TRANSPOSE(cov_fac),temp)
       x(i,1:k) = temp_out + mean(1:k)
    END DO
    DEALLOCATE(temp, temp_out, cov_fac)
!
!print *, "MATMUL in stat_utilities.f90/gen_mvnormal_r(): OK"
    RETURN
  END SUBROUTINE gen_mvnormal_r
!
  SUBROUTINE gen_mvnormal_d(max_n,max_k,n,k,mean,cov,cov_fac_option,x)
!    USE define_types
!
    IMPLICIT NONE
!
    INTEGER, INTENT(in) :: max_n, max_k, n, k, cov_fac_option
    REAL, INTENT(in), DIMENSION(k) :: mean
    REAL, INTENT(in), DIMENSION(max_k, k) :: cov
    REAL, INTENT(inout), DIMENSION(max_n, k) :: x
!
    INTEGER :: info, i, j
    REAL, DIMENSION(:), ALLOCATABLE :: temp, temp_out
    REAL, DIMENSION(:,:), ALLOCATABLE :: cov_fac
!
    EXTERNAL dpotrf
!
    ALLOCATE(cov_fac(k,k))
    cov_fac = cov(1:k,1:k)
    DO i = 1, k
    DO j = 1, i-1
       cov_fac(i,j) = 0.0
    END DO
    END DO
!
    IF(cov_fac_option .NE. 0) CALL dpotrf("U",k,cov_fac,k,info)
!
    ALLOCATE(temp(k))
    ALLOCATE(temp_out(k))
    DO i = 1, n
       CALL gen_normal(0.0,1.0,k,temp)
       temp = MATMUL(TRANSPOSE(cov_fac),temp) 
       x(i,1:k) = temp + mean
    END DO
    DEALLOCATE(temp, temp_out, cov_fac)
!print *, "MATMUL in stat_utilities.f90/gen_mvnormal_d(): OK"
!
    RETURN
  END SUBROUTINE gen_mvnormal_d
!**********************************************************************
!  Smoothing functions:
!**********************************************************************
!    Logisitic:
!**********************************************************************
  FUNCTION logistic_sm_r(x, sm_constant, sm_PARAMETER)
!    USE define_types
!
    IMPLICIT NONE
    REAL, INTENT(in) :: x, sm_constant, sm_PARAMETER
    REAL :: logistic_sm_r
!
    logistic_sm_r = 1.0/(1.0 + EXP(-sm_PARAMETER*(x - sm_constant)))
!
    RETURN
  END FUNCTION logistic_sm_r
!
  FUNCTION logistic_sm_d(x, sm_constant, sm_PARAMETER)
!    USE define_types
!
    IMPLICIT NONE
    REAL, INTENT(in) :: x, sm_constant, sm_PARAMETER
    REAL :: logistic_sm_d
!
    logistic_sm_d = 1.0/(1.0 + EXP(-sm_PARAMETER*(x - sm_constant)))
!
    RETURN
  END FUNCTION logistic_sm_d
!**********************************************************************
!    Exponential:
!**********************************************************************
  FUNCTION exponential_sm_r(x, sm_constant, sm_PARAMETER)
!    USE define_types
!
    IMPLICIT NONE
    REAL, INTENT(in) :: x, sm_constant, sm_PARAMETER
    REAL :: exponential_sm_r
!
    exponential_sm_r = 1.0 - EXP(-sm_PARAMETER*(x - sm_constant)*(x - sm_constant))
!
    RETURN
  END FUNCTION exponential_sm_r
!
  FUNCTION exponential_sm_d(x, sm_constant, sm_PARAMETER)
!    USE define_types
!
    IMPLICIT NONE
    REAL, INTENT(in) :: x, sm_constant, sm_PARAMETER
    REAL :: exponential_sm_d
!
    exponential_sm_d = 1.0 - EXP(-sm_PARAMETER*(x - sm_constant)*(x - sm_constant))
!
    RETURN
  END FUNCTION exponential_sm_d
!**********************************************************************
!    Quadratic logarithm:
!**********************************************************************
  FUNCTION quad_log_sm_r(x, sm_constant, sm_PARAMETER)
!    USE define_types
!
    IMPLICIT NONE
    REAL, INTENT(in) :: x, sm_constant, sm_PARAMETER
    REAL :: quad_log_sm_r
!
    quad_log_sm_r = 1.0/(1.0 + EXP(sm_PARAMETER*(x*x - sm_constant*sm_constant)))
!
    RETURN
  END FUNCTION quad_log_sm_r
!
  FUNCTION quad_log_sm_d(x, sm_constant, sm_PARAMETER)
!    USE define_types
!
    IMPLICIT NONE
    REAL, INTENT(in) :: x, sm_constant, sm_PARAMETER
    REAL:: quad_log_sm_d
!
    quad_log_sm_d = 1.0/(1.0 + EXP(sm_PARAMETER*(x*x - sm_constant*sm_constant)))
!
    RETURN
  END FUNCTION quad_log_sm_d
!***********************************************************************
!    Cubic:
!***********************************************************************
  FUNCTION cubic_sm_r(x)
!    USE define_types
!
    IMPLICIT NONE
    REAL, INTENT(in) :: x
    REAL :: cubic_sm_r
!
    IF(ABS(x) .LE. 1.0) THEN
       cubic_sm_r = 0.75*(1.0 - x*x)
    ELSE
       cubic_sm_r = 0.0
    END IF
!
    RETURN
  END FUNCTION cubic_sm_r
!
  FUNCTION cubic_sm_d(x)
!    USE define_types
!
    IMPLICIT NONE
    REAL, INTENT(in) :: x
    REAL :: cubic_sm_d
!
    IF(ABS(x) .LE. 1.0) THEN
       cubic_sm_d = 0.75*(1.0 - x*x)
    ELSE
       cubic_sm_d = 0.0
    END IF
!
    RETURN
  END FUNCTION cubic_sm_d
!************************************************************************
!     Epanechnikov:
!***********************************************************************
  FUNCTION Epanechnikov_r(x)
!    USE define_types
!
    IMPLICIT NONE
    REAL, INTENT(in) :: x
    REAL :: Epanechnikov_r
!
    IF((x*x) .LE. 0.5) THEN
       Epanechnikov_r = 3.0 * (1.0 - x**(2.0/5.0))
       Epanechnikov_r = Epanechnikov_r / (4.0*SQRT(5.0))
    ELSE
       Epanechnikov_r = 0.0
    END IF
!
    RETURN
  END FUNCTION Epanechnikov_r
!
  FUNCTION Epanechnikov_d(x)
!    USE define_types
!
    IMPLICIT NONE
    REAL, INTENT(in) :: x
    REAL :: Epanechnikov_d
!
    IF((x*x) .LE. 0.5) THEN
       Epanechnikov_d = 3.0 * (1.0 - x**(2.0/5.0))
       Epanechnikov_d = Epanechnikov_d / (4.0*SQRT(5.0))
    ELSE
       Epanechnikov_d = 0.0
    END IF
!
    RETURN
  END FUNCTION Epanechnikov_d
!***********************************************************************
!      Quartic:
!***********************************************************************
  FUNCTION quartic_sm_r(x)
!    USE define_types
!
    IMPLICIT NONE
    REAL, INTENT(in) :: x
    REAL :: quartic_sm_r
!
    IF(ABS(x) .LE. 1.0) THEN
       quartic_sm_r = 0.9375*(1.0 - x*x)*(1.0 - x*x)
    ELSE
       quartic_sm_r = 0.0
    END IF
!
    RETURN
  END FUNCTION quartic_sm_r
!
  FUNCTION quartic_sm_d(x)
!    USE define_types
!
    IMPLICIT NONE
    REAL, INTENT(in) :: x
    REAL :: quartic_sm_d
!
    IF(ABS(x) .LE. 1.0) THEN
       quartic_sm_d = 0.9375*(1.0 - x*x)*(1.0- x*x)
    ELSE
       quartic_sm_d = 0.0
    END IF
!
    RETURN
  END FUNCTION quartic_sm_d
!***********************************************************************
!  Linear regression:
!***********************************************************************
!  Notes: Uses LAPACK subroutines sgetrf and sgetri for computing
!         LU factorization and matrix inversion.  See
!         http://www.netlib.org/lapack/
!***********************************************************************
  SUBROUTINE ls_regr_r(max_n,n,p,x,y,beta_hat)
!    USE define_types
!
    IMPLICIT NONE
    INTEGER, INTENT(in) :: max_n, n, p
    REAL, INTENT(in), DIMENSION(n) :: y
    REAL, INTENT(in), DIMENSION(max_n, p) :: x
    REAL, INTENT(out), DIMENSION(p) :: beta_hat
!
    INTEGER :: info
    INTEGER, DIMENSION(:), ALLOCATABLE :: pivot
    REAL, DIMENSION(:), ALLOCATABLE :: xty
    REAL, DIMENSION(:,:), ALLOCATABLE :: xtx
    REAL, DIMENSION(:), ALLOCATABLE :: work_space
!
    EXTERNAL sgetrf
    EXTERNAL sgetri
!
    ALLOCATE(xtx(p,p))
    ALLOCATE(xty(p))
    xtx = MATMUL(TRANSPOSE(x(1:n,1:p)),x(1:n,1:p))
    xty = MATMUL(TRANSPOSE(x(1:n,1:p)),y(1:n))
!
    ALLOCATE(pivot(p))
    ALLOCATE(work_space(p))
    CALL sgetrf(p,p,xtx,p,pivot,info)
    CALL sgetri(p,xtx,p,pivot,work_space,p,info)
    DEALLOCATE(pivot, work_space)
!
    beta_hat(1:p) = MATMUL(xtx(1:p,1:p),xty(1:p))
!
!print *, "MATMUL (2) in stat_utilities.f90/ls_regr_r(): OK"
    DEALLOCATE(xtx, xty)
    RETURN
  END SUBROUTINE ls_regr_r
!
  SUBROUTINE ls_regr_d(max_n,n,p,x,y,beta_hat)
!    USE define_types
!
    IMPLICIT NONE
    INTEGER, INTENT(in) :: max_n, n, p
    REAL, INTENT(in), DIMENSION(n) :: y
    REAL, INTENT(in), DIMENSION(max_n, p) :: x
    REAL, INTENT(out), DIMENSION(p) :: beta_hat
!
    INTEGER :: info
    INTEGER, DIMENSION(:), ALLOCATABLE :: pivot
    REAL, DIMENSION(:), ALLOCATABLE :: xty, work_space
    REAL, DIMENSION(:,:), ALLOCATABLE :: xtx

!
    EXTERNAL dgetrf
    EXTERNAL dgetri
!
    ALLOCATE(xtx(p,p))
    ALLOCATE(xty(p))
    xtx = MATMUL(TRANSPOSE(x(1:n,1:p)),x(1:n,1:p))
    xty = MATMUL(TRANSPOSE(x(1:n,1:p)),y(1:n))
!
    ALLOCATE(pivot(p))
    ALLOCATE(work_space(p))
    CALL dgetrf(p,p,xtx,p,pivot,info)
    CALL dgetri(p,xtx,p,pivot,work_space,p,info)
    DEALLOCATE(pivot, work_space)
!
    beta_hat(1:p) = MATMUL(xtx(1:p,1:p),xty(1:p))
!
    DEALLOCATE(xtx, xty)
    RETURN
  END SUBROUTINE ls_regr_d
!
  SUBROUTINE ls_regr_rmv(max_n,n,k,p,x,y,beta_dim,beta_hat)
!    USE define_types
!
    IMPLICIT NONE
    INTEGER, INTENT(in) :: max_n, n, k, beta_dim, p
    REAL, INTENT(in), DIMENSION(max_n, p) :: x
    REAL, INTENT(in), DIMENSION(max_n, k) :: y
    REAL, INTENT(inout), DIMENSION(beta_dim, k) :: beta_hat
!
    INTEGER :: info
    INTEGER, DIMENSION(:), ALLOCATABLE :: pivot
    REAL, DIMENSION(:), ALLOCATABLE :: work_space
    REAL, DIMENSION(:,:), ALLOCATABLE :: xtx, xty
!
    EXTERNAL sgetrf
    EXTERNAL sgetri
!
    ALLOCATE(xtx(p,p))
    ALLOCATE(xty(p,k))
    xtx = MATMUL(TRANSPOSE(x(1:n,1:p)),x(1:n,1:p))
    xty = MATMUL(TRANSPOSE(x(1:n,1:p)),y(1:n,1:k))
!
    ALLOCATE(pivot(p))
    ALLOCATE(work_space(p))
    CALL sgetrf(p,p,xtx,p,pivot,info)
    CALL sgetri(p,xtx,p,pivot,work_space,p,info)
    DEALLOCATE(pivot, work_space)
!
    beta_hat(1:p,1:k) = MATMUL(xtx(1:p,1:p),xty(1:p,1:k))
!print *, "MATMUL (3) in stat_utilities.f90/ls_regr_rmv(): OK"
!
    DEALLOCATE(xtx, xty)
    RETURN
!
  END SUBROUTINE ls_regr_rmv
!
  SUBROUTINE ls_regr_dmv(max_n,n,k,p,x,y,beta_dim,beta_hat)
!    USE define_types
!
    IMPLICIT NONE
    INTEGER, INTENT(in) :: max_n, n, k, beta_dim, p
    REAL, INTENT(in), DIMENSION(max_n, p) :: x
    REAL, INTENT(in), DIMENSION(max_n, k) :: y
    REAL, INTENT(inout), DIMENSION(beta_dim, k) :: beta_hat
!
    INTEGER :: info
    INTEGER, DIMENSION(:), ALLOCATABLE :: pivot
    REAL, DIMENSION(:), ALLOCATABLE :: work_space
    REAL, DIMENSION(:,:), ALLOCATABLE :: xtx, xty
!
    EXTERNAL dgetrf
    EXTERNAL dgetri
!
    ALLOCATE(xtx(p,p))
    ALLOCATE(xty(p,k))
    xtx = MATMUL(TRANSPOSE(x(1:n,1:p)),x(1:n,1:p))
    xty = MATMUL(TRANSPOSE(x(1:n,1:p)),y(1:n,1:k))
!
    ALLOCATE(pivot(p))
    ALLOCATE(work_space(p))
    CALL dgetrf(p,p,xtx,p,pivot,info)
    CALL dgetri(p,xtx,p,pivot,work_space,p,info)
    DEALLOCATE(pivot, work_space)
!
    beta_hat(1:p,1:k) = MATMUL(xtx,xty)
!print *, "MATMUL (2) in stat_utilities.f90/ls_regr_dmv(): OK"
!
    DEALLOCATE(xtx,xty)
    RETURN
!
  END SUBROUTINE ls_regr_dmv
!
END MODULE stat_utilities

MODULE stat_interfaces
!***********************************************************************
!  FORTRAN 90 module containing interfaces for statistics utilities in
!  stat_utilities.f90.
!
!  Created: 03/01/03 JLH
!***********************************************************************
  INTERFACE
     SUBROUTINE gen_uniform_r(a,b,n,x)
 !     USE define_types
       IMPLICIT NONE
       INTEGER, INTENT(in) :: n
       REAL, INTENT(in) :: a, b
       REAL, INTENT(out), DIMENSION(n) :: x
     END SUBROUTINE gen_uniform_r
  END INTERFACE
!
  INTERFACE
     SUBROUTINE gen_uniform_d(a,b,n,x)
!      USE define_types
       IMPLICIT NONE
       INTEGER, INTENT(in) :: n
       REAL, INTENT(in) :: a, b
       REAL, INTENT(out), DIMENSION(n) :: x
     END SUBROUTINE gen_uniform_d
  END INTERFACE
!
  INTERFACE
     SUBROUTINE gen_normal_r(mean,variance,n,x)
!      USE define_types
       IMPLICIT NONE
       INTEGER, INTENT(in) :: n
       REAL, INTENT(in) :: mean, variance
       REAL, INTENT(out), DIMENSION(n) :: x
     END SUBROUTINE gen_normal_r
  END INTERFACE
!
  INTERFACE
     SUBROUTINE gen_normal_d(mean,variance,n,x)
!      USE define_types
       IMPLICIT NONE
       INTEGER, INTENT(in) :: n
       REAL,INTENT(in) :: mean, variance
       REAL, INTENT(out), DIMENSION(n) :: x
     END SUBROUTINE gen_normal_d
  END INTERFACE
!
  INTERFACE
     SUBROUTINE gen_mvnormal_r(max_n,max_k,n,k,mean,cov,cov_fac_option,x)
!      USE define_types
       IMPLICIT NONE
       INTEGER, INTENT(in) :: max_n, max_k, n, k, cov_fac_option
       REAL, INTENT(in), DIMENSION(k) :: mean
       REAL, INTENT(in), DIMENSION(max_k, k) :: cov
       REAL, INTENT(out), DIMENSION(max_n, k) :: x
       EXTERNAL spotrf
     END SUBROUTINE gen_mvnormal_r
  END INTERFACE
!
  INTERFACE
     SUBROUTINE gen_mvnormal_d(max_n,max_k,n,k,mean,cov,cov_fac_option,x)
!      USE define_types
       IMPLICIT NONE
       INTEGER, INTENT(in) :: max_n, max_k, n, k, cov_fac_option
       REAL, INTENT(in), DIMENSION(k) :: mean
       REAL, INTENT(in), DIMENSION(max_k, k) :: cov
       REAL, INTENT(inout), DIMENSION(max_n, k) :: x
       EXTERNAL dpotrf
     END SUBROUTINE gen_mvnormal_d
  END INTERFACE
!
  INTERFACE
     FUNCTION logistic_sm_r(x, sm_constant, sm_PARAMETER)
!      USE define_types
       IMPLICIT NONE
       REAL, INTENT(in) :: sm_constant, sm_PARAMETER, x
       REAL :: logistic_sm_r
     END FUNCTION logistic_sm_r
  END INTERFACE
!
  INTERFACE
     FUNCTION logistic_sm_d(x, sm_constant, sm_PARAMETER)
!      USE define_types
       IMPLICIT NONE
       REAL, INTENT(in) :: sm_constant, sm_PARAMETER, x
       REAL :: logistic_sm_d
     END FUNCTION logistic_sm_d
  END INTERFACE
!
  INTERFACE
     FUNCTION exponential_sm_r(x, sm_constant, sm_PARAMETER)
!      USE define_types
       IMPLICIT NONE
       REAL, INTENT(in) :: sm_constant, sm_PARAMETER, x
       REAL :: exponential_sm_r
     END FUNCTION exponential_sm_r
  END INTERFACE
!
  INTERFACE
     FUNCTION exponential_sm_d(x, sm_constant, sm_PARAMETER)
!      USE define_types
       IMPLICIT NONE
       REAL, INTENT(in) :: sm_constant, sm_PARAMETER, x
       REAL :: exponential_sm_d
     END FUNCTION exponential_sm_d
  END INTERFACE
!
  INTERFACE
     FUNCTION quad_log_sm_r(x, sm_constant, sm_PARAMETER)
!      USE define_types
       IMPLICIT NONE
       REAL, INTENT(in) :: sm_constant, sm_PARAMETER, x
       REAL :: quad_log_sm_r
     END FUNCTION quad_log_sm_r
  END INTERFACE
!
  INTERFACE
     FUNCTION quad_log_sm_d(x, sm_constant, sm_PARAMETER)
!      USE define_types
       IMPLICIT NONE
       REAL, INTENT(in) :: sm_constant, sm_PARAMETER, x
       REAL :: quad_log_sm_d
     END FUNCTION quad_log_sm_d
  END INTERFACE
!
  INTERFACE
     FUNCTION cubic_sm_r(x)
!      USE define_types
       IMPLICIT NONE
       REAL, INTENT(in) :: x
       REAL :: cubic_sm_r
     END FUNCTION cubic_sm_r
  END INTERFACE
!
  INTERFACE
     FUNCTION cubic_sm_d(x)
!      USE define_types
       IMPLICIT NONE
       REAL, INTENT(in) :: x
       REAL :: cubic_sm_d
     END FUNCTION cubic_sm_d
  END INTERFACE
!
  INTERFACE
     FUNCTION Epanechnikov_r(x)
!      USE define_types
       IMPLICIT NONE
       REAL, INTENT(in) :: x
       REAL :: Epanechnikov_r
     END FUNCTION Epanechnikov_r
  END INTERFACE
!
  INTERFACE
     FUNCTION Epanechnikov_d(x)
!      USE define_types
       IMPLICIT NONE
       REAL, INTENT(in) :: x
       REAL :: Epanechnikov_d
     END FUNCTION Epanechnikov_d
  END INTERFACE
!
  INTERFACE
     FUNCTION quartic_sm_r(x)
!      USE define_types
       IMPLICIT NONE
       REAL, INTENT(in) :: x
       REAL :: quartic_sm_r
     END FUNCTION quartic_sm_r
  END INTERFACE
!
  INTERFACE
     FUNCTION quartic_sm_d(x)
!      USE define_types
       IMPLICIT NONE
       REAL, INTENT(in) :: x
       REAL :: quartic_sm_d
     END FUNCTION quartic_sm_d
  END INTERFACE
!
  INTERFACE
     SUBROUTINE ls_regr_r(max_n,n,p,x,y,beta_hat)
!      USE define_types
       IMPLICIT NONE
       INTEGER, INTENT(in) :: max_n, n, p
       REAL, INTENT(in), DIMENSION(n) :: y
       REAL, INTENT(in), DIMENSION(max_n, p) :: x
       REAL, INTENT(out), DIMENSION(p) :: beta_hat
       EXTERNAL sgetrf
       EXTERNAL sgetri
     END SUBROUTINE ls_regr_r
  END INTERFACE
!
  INTERFACE
     SUBROUTINE ls_regr_d(max_n,n,p,x,y,beta_hat)
!      USE define_types
       IMPLICIT NONE
       INTEGER, INTENT(in) :: max_n, n, p
       REAL, INTENT(in), DIMENSION(n) :: y
       REAL, INTENT(in), DIMENSION(max_n, p) :: x
       REAL, INTENT(out), DIMENSION(p) :: beta_hat
       EXTERNAL dgetrf
       EXTERNAL dgetri
     END SUBROUTINE ls_regr_d
  END INTERFACE
!
  INTERFACE
     SUBROUTINE ls_regr_rmv(max_n,n,k,p,x,y,beta_dim,beta_hat)
!      USE define_types
       IMPLICIT NONE
       INTEGER, INTENT(in) :: max_n, n, k, beta_dim, p
       REAL, INTENT(in), DIMENSION(max_n, p) :: x
       REAL, INTENT(in), DIMENSION(max_n, k) :: y
       REAL, INTENT(inout), DIMENSION(beta_dim, k) :: beta_hat
       EXTERNAL sgetrf
       EXTERNAL sgetri
     END SUBROUTINE ls_regr_rmv
  END INTERFACE
!
  INTERFACE
     SUBROUTINE ls_regr_dmv(max_n,n,k,p,x,y,beta_dim,beta_hat)
!      USE define_types
       IMPLICIT NONE
       INTEGER, INTENT(in) :: max_n, n, k, beta_dim, p
       REAL, INTENT(in), DIMENSION(max_n, p) :: x
       REAL, INTENT(in), DIMENSION(max_n, k) :: y
       REAL, INTENT(inout), DIMENSION(beta_dim, k) :: beta_hat
       EXTERNAL dgetrf
       EXTERNAL dgetri
     END SUBROUTINE ls_regr_dmv
  END INTERFACE
!
END MODULE stat_interfaces

MODULE nlts_utilities
!***********************************************************************
!   FORTRAN 90 module containing commonly used nonlinear time series
!   statistical computations.
!
!   Created: 03/12/03 JLH
!
!   Contents:
!        Subroutines for generating nonlinear vector realizations:
!            get_mvbilinear_params, gen_mvbilinear, get_mvTAR_params,
!            gen_mvTAR, get_mvSTAR_params, gen_mvSTAR,
!            get_mvEXPAR_params, gen_mvEXPAR
!***********************************************************************
! USE define_types
! USE stat_utilities
!
  IMPLICIT NONE
CONTAINS
!***********************************************************************
!  Generate vector bilinear:
!***********************************************************************
!  Get coefficients:
!***********************************************************************
  SUBROUTINE get_mvbilinear_params(max_k,max_p,max_q,file_label,k,p,q,  &
       cp_option,AR_coefs,MA_coefs,cp_dim,cp_coefs)
!***********************************************************************
!  FORTRAN90 subroutine to get parameter values for generating 
!  multivariate bilinear time series.
!***********************************************************************
!    USE define_types
    IMPLICIT NONE
!
    INTEGER, INTENT(in) :: max_k, max_p, max_q, file_label, k
    INTEGER, INTENT(out) :: p, q, cp_option, cp_dim
    REAL, INTENT(out), DIMENSION(max_k, (max_k*max_p)**2) :: cp_coefs
    REAL, INTENT(out), DIMENSION(max_p, max_k, k) :: AR_coefs
    REAL, INTENT(out), DIMENSION(max_q, max_k, k) :: MA_coefs
!
    INTEGER :: lag, dim1, dim2
!
    WRITE(file_label,400) k
    WRITE(*,*) "Enter the AR order (p) and MA order (q)."
    WRITE(*,*) "p = ?"
    READ(*,*) p
    WRITE(*,*) "q = ?"
    READ(*,*) q
    WRITE(file_label,500) p, q
!
    IF(p > 0) THEN
       WRITE(*,*) "Enter the AR coefficients, row by row."
       WRITE(file_label,510) 
       DO lag = 1, p
          WRITE(*,520) lag
          WRITE(file_label,520) lag
          DO dim1 = 1, k
             READ(*,*) (AR_coefs(lag,dim1,dim2), dim2 = 1, k)
             WRITE(file_label,*) (AR_coefs(lag,dim1,dim2), dim2 = 1, k)
          END DO
       END DO
    END IF
!
    IF(q > 0) THEN
       WRITE(*,*) "Enter the MA coefficients, row by row."
       WRITE(file_label,530) 
       DO lag = 1, q
          WRITE(*,520) lag
          WRITE(file_label,520) lag
          DO dim1 = 1, k
             READ(*,*) (MA_coefs(lag,dim1,dim2), dim2 = 1, k)
             WRITE(file_label,*) (MA_coefs(lag,dim1,dim2), dim2 = 1, k)
          END DO
       END DO
    END IF
!
    WRITE(*,*) "What type of cross-product terms should be included?"
    WRITE(*,*) " 0 => no cross-product terms,"
    WRITE(*,*) " 1 => time series realizations with errors,"
    WRITE(*,*) " 2 => squared terms and interactions of time series, or"
    WRITE(*,*) " 3 => squared and cross-products of error terms."
    READ(*,*) cp_option
    WRITE(file_label,*)
!
    IF(cp_option == 0) THEN
       cp_dim = 1
       WRITE(file_label,540)
    ELSE IF(cp_option == 1) THEN
       WRITE(file_label,550)
       cp_dim = p*q*k*k
    ELSE IF(cp_option == 2) THEN
       WRITE(file_label,560)
       cp_dim = (p*k)*(p*k + 1)/2
    ELSE IF(cp_option == 3) THEN
       WRITE(file_label,570)
       cp_dim = (q*k)*(q*k + 1)/2
    END IF
    WRITE(file_label,*)
!
    WRITE(*,*) "Enter the coefficients for the second order terms (row by row)."
    WRITE(file_label,580)
    DO dim1 = 1, k
       READ(*,*) (cp_coefs(dim1,dim2), dim2 = 1, cp_dim)
       WRITE(file_label,*) (cp_coefs(dim1,dim2), dim2 = 1, cp_dim)
    END DO
!
400 FORMAT("Vector bilinear process of dimension ",I1,".")
500 FORMAT("AR order (p) = ",I2,": MA order (q) = ",I2,".")
510 FORMAT("The AR coefficients (row by row) are:")
520 FORMAT("Lag ",I2,":")
530 FORMAT("The MA coefficients (row by row) are:")
540 FORMAT("There are no cross-product terms.")
550 FORMAT("Cross-product terms were the time series with the errors.")
560 FORMAT("Cross-product terms were squared terms and interactions of the time series.")
570 FORMAT("Cross-product terms were squared errors and cross-products of errors.")
580 FORMAT("The second-order term coefficients are (row by row):")
!
    RETURN
  END SUBROUTINE get_mvbilinear_params
!***********************************************************************
!   Generate realization:
!***********************************************************************
  SUBROUTINE gen_mvbilinear(max_n,max_k,max_p,max_q,n,k,p,q,AR_coefs,   &
       MA_coefs,cp_option,cp_dim,cp_coefs,cov_fac_option,cov,extra,x)
!   USE define_types
!   USE f90_utilities, ONLY : vech
!   USE stat_utilities
!
    IMPLICIT NONE
    INTEGER, INTENT(in) :: max_n, max_k, max_p, max_q
    INTEGER, INTENT(in) :: n, k, p, q, cp_option, cp_dim
    INTEGER, INTENT(in) :: cov_fac_option, extra
    REAL, INTENT(in), DIMENSION(max_k, k) :: cov
    REAL, INTENT(in), DIMENSION(max_k, cp_dim) :: cp_coefs
    REAL, INTENT(in), DIMENSION(max_p, max_k, k) :: AR_coefs
    REAL, INTENT(in), DIMENSION(max_q, max_k, k) :: MA_coefs
    REAL, INTENT(inout), DIMENSION(max_n, k) :: x
!
    REAL, DIMENSION(max_n,k) :: eps
    INTEGER :: nn, dim, t, i, j, index1, index2, vech_option, jj, l, dim1, dim2
    REAL, DIMENSION(k) :: work_space
    REAL, DIMENSION((p*k)) :: zv
    REAL, DIMENSION(cp_dim) :: z
    REAL, DIMENSION(1,(p*k)) :: zvt
!
    work_space = 0.0
    eps = 0.0
    nn = n + extra
    dim = max(p,q)*k
    CALL gen_normal(max_n,max_k,nn,k,work_space,cov,cov_fac_option,eps)
!
    x(1:nn,1:k) = eps(1:nn,1:k)
    DO t = 1, nn
!
!  Terms for computing cross-products:
!
       DO i = 1, k
          DO j = 1, MIN(t-1, p)
             index1 = t - j
             index2 = p*(i-1) + j
             IF(cp_option == 1) THEN
                zv(index2) = x(index1,i)
                zvt(1,index2) = eps(index1,i)
                vech_option = 1
             ELSE IF(cp_option == 2) THEN
                zv(index2) = x(t-j,i)
                zvt(1,index2) = zv(p*(i-1) + j)
                vech_option = 0
             ELSE IF(cp_option == 3) THEN
                zv(index2) = eps(index1,i)
                zvt(1,index2) = zv(index2)
                vech_option = 0
             END IF
             CALL vech(zv,zvt,dim,dim,1,dim,p*k,vech_option,z)
          END DO
       END DO
!
!  Linear AR part:
!
       DO j = 1, MIN(t-1, p)
          DO dim1 = 1, k
          DO dim2 = 1, k
             x(t,dim1) = x(t,dim1) + AR_coefs(j,dim1,dim2)*x(t-j,dim2)
          END DO
          END DO
       END DO
!
!  Linear MA part:
!
       DO j = 1, MIN(t-1, q)
          DO dim1 = 1, k
          DO dim2 = 1, k
             x(t,dim1) = x(t,dim1) - MA_coefs(j,dim1,dim2)*eps(t-j,dim2)
          END DO
          END DO
       END DO
!
!  Nonlinear part:
!
       DO i = 1, k
       DO j = 1, cp_dim
          x(t,i) = x(t,i) + cp_coefs(i,j)*z(j)
       END DO
       END DO
    END DO
!
    x(1:n,1:k) = x((extra+1):(extra+n),1:k)
!
    RETURN
  END SUBROUTINE gen_mvbilinear
!***********************************************************************
!  Generate vector threshold ARMA:
!***********************************************************************
!  Get coefficients:
!***********************************************************************
  SUBROUTINE get_mvTAR_params(max_k,max_p,max_q,max_no_regimes,       &
       file_label,k,p,q,no_regimes,TAR_AR_coefs,TAR_MA_coefs,         &
       AR_lags,MA_lags,thr_resp_ind,thr_lag,thresholds,TAR_covs)
!***********************************************************************
!  FORTRAN 90 subroutine to get parameters for generating realizations
!  from a multivariate threshold AR process.
!***********************************************************************
!   USE define_types
    IMPLICIT NONE
!
    INTEGER, INTENT(in) :: max_k, max_p, max_q, max_no_regimes 
    INTEGER, INTENT(in) :: file_label, k
    INTEGER, INTENT(out) :: p, q, no_regimes, thr_resp_ind, thr_lag
    INTEGER, INTENT(out), DIMENSION(max_p) :: AR_lags
    INTEGER, INTENT(out), DIMENSION(max_q) :: MA_lags
    REAL, INTENT(out), DIMENSION(max_no_regimes) :: thresholds
    REAL, INTENT(out), DIMENSION(max_no_regimes, max_k, max_k) :: TAR_covs
    REAL, INTENT(out), DIMENSION(max_no_regimes, max_p, max_k, max_k) :: TAR_AR_coefs
    REAL, INTENT(out), DIMENSION(max_no_regimes, max_q, max_k, max_k) :: TAR_MA_coefs
!
    INTEGER :: reg_ct, dim1, dim2, info, j
    REAL, DIMENSION(k, k) :: cov
!
    WRITE(file_label,400) k
    WRITE(*,*) "How many regimes in the TAR model?"
    READ(*,*) no_regimes
    WRITE(file_label,500) no_regimes
    WRITE(file_label,*)
!
    WRITE(*,*) "Enter the AR order (p) and MA order (q)."
    WRITE(*,*) "p = ?"
    READ(*,*) p
    WRITE(*,*) "q = ?"
    READ(*,*) q
    WRITE(file_label,510) p, q
    WRITE(file_label,*)
!
    DO reg_ct = 1, no_regimes
       WRITE(*,520) reg_ct
       WRITE(file_label,520) reg_ct
       DO dim1 = 1,k
          READ(*,*) (cov(dim1,dim2), dim2 = 1, k)
          WRITE(file_label,*) (cov(dim1,dim2), dim2 = 1, k)
       END DO
       TAR_covs(reg_ct,1:k,1:k) = cov(1:k,1:k)
       WRITE(file_label,*)
    END DO
!
    IF(p > 0) THEN
       WRITE(*,*) "What are the AR lags?"
       READ(*,*) (AR_lags(j), j = 1, p)
       WRITE(file_label,525,ADVANCE="NO") 
       WRITE(file_label,*) (AR_lags(j), j = 1, p)
       DO reg_ct = 1, no_regimes
          WRITE(*,530) reg_ct
          WRITE(file_label,540) reg_ct
          DO j = 1, p
             WRITE(*,550) j
             DO dim1 = 1,k
                READ(*,*) (TAR_AR_coefs(reg_ct,j,dim1,dim2), dim2 = 1, k)
                WRITE(file_label,*) (TAR_AR_coefs(reg_ct,j,dim1,dim2), dim2 = 1, k)
             END DO
          END DO
       END DO
    END IF
!
    IF(q > 0) THEN
       WRITE(*,*) "What are the MA lags?"
       READ(*,*) (MA_lags(j), j = 1, q)
       WRITE(file_label,*)
       WRITE(file_label,555,ADVANCE="NO") 
       WRITE(file_label,*) (MA_lags(j), j = 1, q)
       DO reg_ct = 1, no_regimes
          WRITE(*,560) reg_ct
          WRITE(file_label,570) reg_ct
          DO j = 1, q
             WRITE(*,580) j
             DO dim1 = 1, k
                READ(*,*) (TAR_MA_coefs(reg_ct,j,dim1,dim2), dim2 = 1, k)
                WRITE(file_label,*) (TAR_MA_coefs(reg_ct,j,dim1,dim2), dim2 = 1, k)
             END DO
          END DO
       END DO
    END IF
!
    WRITE(*,590) 
    WRITE(*,*) "Threshold response index?"
    READ(*,*) thr_resp_ind
    WRITE(*,*) "Threshold lag?"
    READ(*,*) thr_lag
    WRITE(file_label,*)
    WRITE(file_label,600) thr_lag, thr_resp_ind
    !
    thresholds(1) = -9.0e15
    thresholds(no_regimes+1) = 9.0e10
    IF(no_regimes > 1) THEN
       WRITE(*,610) no_regimes-1
       READ(*,*) (thresholds(reg_ct+1), reg_ct = 1, no_regimes-1)
       WRITE(file_label,620,ADVANCE="NO")
       WRITE(file_label,*) (thresholds(reg_ct+1), reg_ct = 1, no_regimes-1)
    END IF
!
400 FORMAT("Vector threshold autoregressive moving average model of dimension ",I1,".")
500 FORMAT("The number of regimes in the TAR model is ",I2,".")
510 FORMAT("AR order (p) = ",I2,": MA order (q) = ",I2,".")
520 FORMAT("Covariance matrix of errors (row by row) for regime ",I2,".")
525 FORMAT("AR lags:")
530 FORMAT("Enter the AR matrix of coefficients for regime ",I2,".")
540 FORMAT("The AR matrix of coefficients for regime ",I2,":")
550 FORMAT("Enter the lag ",I2," AR matrix of coefficients, row by row.")
555 FORMAT("MA lags:")
560 FORMAT("Enter the MA matrix of coefficients for regime ",I2,".")
570 FORMAT("The MA matrix of coefficients for regime ",I2,":")
580 FORMAT("Enter the lag ",I2," MA matrix of coefficients, row by row.")
590 FORMAT("Enter the threshold response index and lag in that order.")
600 FORMAT("The threshold variable is X(t -",I2,",",I2,").")
610 FORMAT("Input ",I2," thesholds.")
620 FORMAT("The thresholds are: ")
!
    RETURN
  END SUBROUTINE get_mvTAR_params
!***********************************************************************
!  Generate realization:
!***********************************************************************
  SUBROUTINE gen_mvTAR(max_n,max_k,max_p,max_q,max_no_regimes,          &
       n,k,p,q,no_regimes,TAR_AR_coefs,TAR_MA_coefs,AR_lags,MA_lags,    &
       thr_resp_ind,thr_lag,thresholds,TAR_covs,cov_fac_option,extra,x)
!   USE define_types
!   USE stat_utilities
!
    IMPLICIT NONE
!
    INTEGER, INTENT(in) :: max_n, max_k, max_p, max_q, max_no_regimes
    INTEGER, INTENT(in) :: n, k, p, q, cov_fac_option, extra
    INTEGER, INTENT(in) :: no_regimes, thr_resp_ind, thr_lag
    INTEGER, INTENT(in), DIMENSION(p) :: AR_lags
    INTEGER, INTENT(in), DIMENSION(q) :: MA_lags
    REAL, INTENT(in), DIMENSION(max_no_regimes) :: thresholds
    REAL, INTENT(in), DIMENSION(max_no_regimes, max_p, max_k, k) :: TAR_AR_coefs
    REAL, INTENT(in), DIMENSION(max_no_regimes, max_q, max_k, k) :: TAR_MA_coefs
    REAL, INTENT(in), DIMENSION(max_no_regimes, max_k, k) :: TAR_covs
    REAL, INTENT(out), DIMENSION (max_n, k) :: x
!
    INTEGER :: regime_no, regime, t, j, nn
    REAL :: crit, thr1, thr2
    REAL, DIMENSION(max_n, k) :: eps
    REAL, DIMENSION(k) :: work_space
    REAL, DIMENSION(max_k, k) :: cov
    REAL, DIMENSION(no_regimes+2,2*n,k) :: all_eps
!
    nn = n + extra
    work_space = 0.0
    DO regime_no = 1, no_regimes
       cov(1:k,1:k) = TAR_covs(regime_no,1:k,1:k)
       CALL gen_normal(max_n,max_k,nn,k,work_space,cov,cov_fac_option,eps)
       all_eps(regime_no,1:nn,1:k) = eps(1:nn,1:k)
    END DO
!
    work_space = 0.0
    x = 0.0
    DO t = thr_lag+1, nn
       crit = x(t-thr_lag, thr_resp_ind)
       regime = 1
       DO regime_no = 1, no_regimes
          thr1 = thresholds(regime_no)
          thr2 = thresholds(regime_no+1)
          IF(crit .GE. thr1 .AND. crit .LT. thr2) THEN
             regime = regime_no 
             GO TO 10
          END IF
       END DO
10     CONTINUE
!
       x(t,1:k) = all_eps(regime,t,1:k)
       DO j = 1, MIN(t-1, p)
          work_space = MATMUL(x(t-AR_lags(j),1:k),TAR_AR_coefs(regime,j,1:k,1:k))
          x(t,1:k) = x(t,1:k) + work_space
       END DO
       DO j = 1, MIN(t-1, q)
          work_space = MATMUL(all_eps(regime,t-MA_lags(j),1:k),TAR_MA_coefs(regime,j,1:k,1:k))
          x(t,1:k) = x(t,1:k) + work_space
       END DO
    END DO
!
    x(1:n,1:k) = x((extra+1):nn,1:k)
    RETURN
!
  END SUBROUTINE gen_mvTAR
!***********************************************************************
!  Generate vector threshold STAR:
!***********************************************************************
!  Get coefficients:
!***********************************************************************
  SUBROUTINE get_mvSTAR_params(max_k,max_p,max_no_regimes,file_label, &
       k,p,sm_FUNCTION,sm_const,sm_param,STAR_AR_coefs,AR_lags,       &
       thr_resp_ind,thr_lag)
!***********************************************************************
!  FORTRAN 90 subroutine to get parameters for generating realizations
!  from a multivariate smoothed threshold ARMA process.
!***********************************************************************
!    USE define_types
    IMPLICIT NONE
!
    INTEGER, INTENT(in) :: max_k, max_p, max_no_regimes
    INTEGER, INTENT(in) :: file_label, k
    INTEGER, INTENT(out) :: p, sm_FUNCTION, thr_resp_ind, thr_lag
    INTEGER, INTENT(out), DIMENSION(max_p) :: AR_lags
    REAL, INTENT(out) :: sm_param, sm_const
    REAL, INTENT(out), DIMENSION(max_no_regimes, max_p, max_k, max_k) :: STAR_AR_coefs
!
    INTEGER :: reg_ct, dim1, dim2, info, j
    REAL, DIMENSION(k,k) :: cov
!
    WRITE(file_label,300) k
!
    WRITE(*,*) "What function should be used for smoothing?"
    WRITE(*,*) "  (1 = logistic, 2 = exponential, 3 = quadratic log)"
    READ(*,*) sm_FUNCTION
    WRITE(*,*) "What is the value for the smoothing function constant term?"
    READ(*,*) sm_const
    WRITE(*,*) "What is the value for the smoothing function parameter?"
    READ(*,*) sm_param
    IF(sm_FUNCTION == 1) THEN
       WRITE(file_label,400)
    ELSE IF(sm_FUNCTION == 2) THEN
       WRITE(file_label,410)
    ELSE IF(sm_FUNCTION == 3) THEN
       WRITE(file_label,420)
    END IF
    WRITE(file_label,430) sm_const
    WRITE(file_label,440) sm_param
    WRITE(file_label,*)
!
    WRITE(*,*) "Enter the AR order."
    WRITE(*,*) "p = ?"
    READ(*,*) p
    WRITE(file_label,510) p
    WRITE(file_label,*)
!
    IF(p > 0) THEN
       WRITE(*,*) "What are the AR lags?"
       READ(*,*) (AR_lags(j), j = 1, p)
       WRITE(file_label,525,ADVANCE="NO") 
       WRITE(file_label,*) (AR_lags(j), j = 1, p)
       DO reg_ct = 1, 2
          WRITE(*,530) reg_ct
          WRITE(file_label,540) reg_ct
          DO j = 1, p
             WRITE(*,550) j
             DO dim1 = 1,k
                READ(*,*) (STAR_AR_coefs(reg_ct,j,dim1,dim2), dim2 = 1, k)
                WRITE(file_label,*) (STAR_AR_coefs(reg_ct,j,dim1,dim2), dim2 = 1, k)
             END DO
          END DO
       END DO
    END IF
!
    WRITE(*,590) 
    WRITE(*,*) "Threshold response index?"
    READ(*,*) thr_resp_ind
    WRITE(*,*) "Threshold lag?"
    READ(*,*) thr_lag
    WRITE(file_label,*)
    WRITE(file_label,600) thr_lag, thr_resp_ind
!
300 FORMAT("Smoothed threshold autoregressive process of dimension ",I1,".")
!
400 FORMAT("The function used for smoothing was the logistic function.")
410 FORMAT("The function used for smoothing was the exponential function.")
420 FORMAT("The function used for smoothing was the quadratic log function.")
430 FORMAT("The smoothing function constant term is ",F8.4,".")
440 FORMAT("The smoothing function parameter is",F8.4,".")
!
510 FORMAT("AR order (p) = ",I2,".")
525 FORMAT("AR lags:")
530 FORMAT("Enter the AR matrix of coefficients for regime ",I2,".")
540 FORMAT("The AR matrix of coefficients for regime ",I2,":")
550 FORMAT("Enter the lag ",I2," AR matrix of coefficients, row by row.")
590 FORMAT("Enter the threshold response index and lag in that order.")
600 FORMAT("The threshold variable is X(t -",I2,",",I2,").")
!
    RETURN
  END SUBROUTINE get_mvSTAR_params
!***********************************************************************
!  Generate realization:
!***********************************************************************
  SUBROUTINE gen_mvSTAR(max_n,max_k,max_p,max_no_regimes,n,k,p,         &
       sm_function,sm_const,sm_param,STAR_AR_coefs,AR_lags,             &
       thr_resp_ind,thr_lag,cov,cov_fac_option,extra,x)
!    USE define_types
!    USE stat_utilities
!
    IMPLICIT NONE
!
    INTEGER, INTENT(in) :: max_n, max_k, max_p, max_no_regimes
    INTEGER, INTENT(in) :: n, k, p, cov_fac_option, extra
    INTEGER, INTENT(in) :: sm_FUNCTION, thr_resp_ind, thr_lag
    INTEGER, INTENT(in), DIMENSION(p) :: AR_lags
    REAL, INTENT(in) :: sm_const, sm_param
    REAL, INTENT(in), DIMENSION(max_no_regimes, max_p, max_k, k) :: STAR_AR_coefs
    REAL, INTENT(in), DIMENSION(max_k, k) :: cov
    REAL, INTENT(out), DIMENSION (max_n, k) :: x
!
    INTEGER :: nn, regime_no, regime, t, j
    REAL :: s
    REAL, DIMENSION(max_n, k) :: eps
    REAL, DIMENSION(k) :: work_space1, work_space2
!
    nn = n + extra
    work_space1 = 0.0
    eps = 0.0
    CALL gen_normal(max_n,max_k,nn,k,work_space1,cov,cov_fac_option,eps)
!
    x(1:nn,1:k) = eps(1:nn,1:k)
    DO t = thr_lag+1, nn
       work_space1 = 0.0
       work_space2 = 0.0
!
       DO j = 1, MIN(t-1, p)
          work_space1 = MATMUL(x(t-AR_lags(j),1:k),STAR_AR_coefs(1,j,1:k,1:k))
       END DO
       DO j = 1, MIN(t-1, p)
          work_space2 = MATMUL(x(t-AR_lags(j),1:k),STAR_AR_coefs(2,j,1:k,1:k))
       END DO
!
       s = 0.0
       IF(sm_FUNCTION == 1) THEN
!!          s = logistic_sm(x(t-thr_lag,thr_resp_ind),sm_const,sm_param)
       ELSE IF(sm_FUNCTION == 2) THEN
!!          s = exponential_sm(x(t-thr_lag,thr_resp_ind),sm_const,sm_param)
       ELSE IF(sm_FUNCTION == 3) THEN
!!          s = quad_log_sm(x(t-thr_lag,thr_resp_ind),sm_const,sm_param)
       END IF
!
       x(t,1:k) = work_space1*(1.0 - s) + work_space2*s + eps(t,1:k)
    END DO
!
    x(1:n,1:k) = x((extra+1):nn,1:k)
!
    RETURN
  END SUBROUTINE gen_mvSTAR
!***********************************************************************
!  Generate vector exponential AR:
!***********************************************************************
!  Get coefficients:
!***********************************************************************
  SUBROUTINE get_mvEXPAR_params(max_k,max_p,file_label,k,p,AR_cons,      &
       AR_coefs,expar_coefs,d)
!***********************************************************************
!  FORTRAN90 subroutine to get parameters for multivariate exponential
!  autoregressive process.
!***********************************************************************
!    USE define_types
    IMPLICIT NONE
!
    INTEGER, INTENT(in) :: max_k, max_p, file_label, k
    INTEGER, INTENT(out) :: p, d
    REAL, INTENT(out), DIMENSION(max_p, max_k, max_k) :: AR_cons, AR_coefs
    REAL, INTENT(out), DIMENSION(max_p, max_k*max_k, max_k) :: expar_coefs
!
    INTEGER :: ts_lag, dim1, dim2, i
!
    WRITE(file_label,400) k
    WRITE(*,*) "Enter the AR order (p)."
    READ(*,*) p
    WRITE(file_label,500) p
    WRITE(file_label,*)
!
    IF(p > 0) THEN
       WRITE(*,*) "Enter the AR intercept terms (row by row)."
       WRITE(file_label,510)
       DO ts_lag = 1, p
          WRITE(*,520) ts_lag
          WRITE(file_label,520) ts_lag
          DO dim1 = 1, k
             READ(*,*) (AR_cons(ts_lag,dim1,dim2), dim2 = 1, k)
             WRITE(file_label,*) (AR_cons(ts_lag,dim1,dim2), dim2 = 1, k)
          END DO
       END DO
!
       WRITE(*,*) "Enter the AR coefficients (row by row)."
       WRITE(file_label,*)
       WRITE(file_label,530)
       DO ts_lag = 1, p
          WRITE(*,520) ts_lag
          WRITE(file_label,520) ts_lag
          DO dim1 = 1, k
             READ(*,*) (AR_coefs(ts_lag,dim1,dim2), dim2 = 1, k)
             WRITE(file_label,*) (AR_coefs(ts_lag,dim1,dim2), dim2 = 1, k)
          END DO
       END DO
!
       WRITE(*,*) "Enter the delay for the exponential term."
       READ(*,*) d
       WRITE(file_label,550) d

       WRITE(*,*) "Enter the exponential parameters (row by row)."
       WRITE(file_label,*)
       WRITE(file_label,540)
       DO ts_lag = 1, p
          WRITE(*,520) ts_lag
          WRITE(file_label,520) ts_lag
          DO dim1 = 1, k
             WRITE(*,525) dim1
             WRITE(file_label,525) dim1
             READ(*,*) (expar_coefs(ts_lag,dim2,dim1), dim2 = 1, k*k)
             WRITE(file_label,*) (expar_coefs(ts_lag,dim2,dim1), dim2 = 1, k*k)
          END DO
       END DO
!
    END IF
!
400 FORMAT("Vector exponential autoregressive process of dimension ",I1,".")
500 FORMAT("The AR order (p) is ",I2,".")
510 FORMAT("The AR constant terms (row by row):")
520 FORMAT("Lag ",I2,":")
525 FORMAT("Time series component ",I1,":")
530 FORMAT("The AR coefficients (row by row) are:")
540 FORMAT("The exponential parameters (row by row) are:")
550 FORMAT("Delay for the exponential variable: ",I1,".")
!
    RETURN
  END SUBROUTINE get_mvEXPAR_params
!***********************************************************************
!  Generate realization:
!***********************************************************************
  SUBROUTINE gen_mvEXPAR(max_n,max_k,max_p,n,k,p,AR_cons,AR_coefs,      &
       expar_coefs,d,cov,cov_fac_option,extra,x)
!    USE define_types
!    USE stat_utilities
!
    IMPLICIT NONE
    INTEGER, INTENT(in) :: max_n, max_k, max_p, n, k, p, d
    INTEGER, INTENT(in) :: cov_fac_option, extra
    REAL, INTENT(in), DIMENSION(max_k, k) :: cov
    REAL, INTENT(in), DIMENSION(max_p, max_k, k) :: AR_cons, AR_coefs
    REAL, INTENT(in), DIMENSION(max_p, max_k*max_k, k) :: expar_coefs
    REAL, INTENT(out), DIMENSION(max_n, k) :: x
!
    INTEGER :: nn, s, t, j, dim1, dim2, i, ii
    REAL, DIMENSION(max_n, k) :: eps
    REAL, DIMENSION(k) :: work_space, temp
    REAL, DIMENSION(k, k) :: cc
    REAL, DIMENSION(p*k, k) :: eterm
    REAL, DIMENSION(p, k, k) :: coefs
!
    nn = n + extra
    work_space = 0.0
    eterm = 0.0
    temp = 0.0
    eps = 0.0
    s = max(p,d)
    CALL gen_normal(max_n,max_k,nn,k,work_space,cov,cov_fac_option,eps)
!
    x(1:nn,1:k) = eps(1:nn,1:k)
    DO t = s+1, nn
       coefs = 0.0
       temp(1:k) = x(t-d,1:k)*x(t-d,1:k)
       DO j = 1, p	   
          DO dim1 = 1, k	   
             DO dim2 = 1, k
                i = k*(dim2-1) + 1
                ii = dim2*k
                work_space(1:k) = expar_coefs(j,i:ii,dim1)
                eterm(dim1,dim2) = DOT_PRODUCT(temp,work_space)
             END DO
          END DO
          coefs(j,1:k,1:k) = AR_coefs(j,1:k,1:k)*EXP(eterm(1:k,1:k))
          coefs(j,1:k,1:k) = coefs(j,1:k,1:k) + AR_cons(j,1:k,1:k)
!
          cc(1:k,1:k) = coefs(j,1:k,1:k)
          cc = TRANSPOSE(cc)
!
          work_space(1:k) = MATMUL(x(t-j,1:k),cc)
          x(t,1:k) = x(t,1:k) + work_space(1:k)
       END DO
    END DO
!  
    x(1:n,1:k) = x((extra+1):nn,1:k)
!
    RETURN
  END SUBROUTINE gen_mvEXPAR
!
END MODULE nlts_utilities

MODULE nlts_interfaces
!***********************************************************************
!  FORTRAN 90 module containing interfaces for non-linear time series
!  analysis programs.
!***********************************************************************
  INTERFACE
     SUBROUTINE get_mvbilinear_params(max_k,max_p,max_q,file_label,k,p,q,&
		  cp_option,AR_coefs,MA_coefs,cp_dim,cp_coefs)
!       USE define_types
       IMPLICIT NONE
       INTEGER, INTENT(in) :: max_k, max_p, max_q, file_label, k
       INTEGER, INTENT(out) :: p, q, cp_option, cp_dim
       REAL, INTENT(out), DIMENSION(max_k, (max_p*max_k)**2) :: cp_coefs
       REAL, INTENT(out), DIMENSION(max_p, max_k, k) :: AR_coefs
       REAL, INTENT(out), DIMENSION(max_q, max_k, k) :: MA_coefs
     END SUBROUTINE get_mvbilinear_params
  END INTERFACE
!
  INTERFACE
     SUBROUTINE gen_mvbilinear(max_n,max_k,max_p,max_q,n,k,p,q,AR_coefs, &
          MA_coefs,cp_option,cp_dim,cp_coefs,cov_fac_option,cov,extra,x)
 !      USE define_types
 !      USE f90_utilities, ONLY : vech
 !      USE stat_utilities
       IMPLICIT NONE
       INTEGER, INTENT(in) :: max_n, max_k, max_p, max_q
       INTEGER, INTENT(in) :: n, k, p, q, cp_option, cp_dim
       INTEGER, INTENT(in) :: cov_fac_option, extra
       REAL, INTENT(in), DIMENSION(max_k, k) :: cov
       REAL, INTENT(in), DIMENSION(max_k, cp_dim) :: cp_coefs
       REAL, INTENT(in), DIMENSION(max_p, max_k, k) :: AR_coefs
       REAL, INTENT(in), DIMENSION(max_q, max_k, k) :: MA_coefs
       REAL, INTENT(inout), DIMENSION(max_n, k) :: x
     END SUBROUTINE gen_mvbilinear
  END INTERFACE
!
  INTERFACE
     SUBROUTINE get_mvTAR_params(max_k,max_p,max_q,max_no_regimes,         &
	      file_label,k,p,q,no_regimes,TAR_AR_coefs,TAR_MA_coefs,           &
          AR_lags,MA_lags,thr_resp_ind,thr_lag,thresholds,TAR_covs)
!       USE define_types
       IMPLICIT NONE
       INTEGER, INTENT(in) :: max_k, max_p, max_q, max_no_regimes
       INTEGER, INTENT(in) :: file_label, k
       INTEGER, INTENT(out) :: p, q, no_regimes, thr_resp_ind, thr_lag
       INTEGER, INTENT(out), DIMENSION(max_p) :: AR_lags
       INTEGER, INTENT(out), DIMENSION(max_q) :: MA_lags
       REAL, INTENT(out), DIMENSION(max_no_regimes) :: thresholds
       REAL, INTENT(out), DIMENSION(max_no_regimes, max_k, max_k) :: TAR_covs
       REAL, INTENT(out), DIMENSION(max_no_regimes, max_p, max_k, max_k) :: TAR_AR_coefs
	   REAL, INTENT(out), DIMENSION(max_no_regimes, max_q, max_k, max_k) :: TAR_MA_coefs
     END SUBROUTINE get_mvTAR_params
  END INTERFACE
!
!!INTERFACE
!!   SUBROUTINE gen_mvTAR(max_n,max_k,max_p,max_q,max_no_regimes,          &
!!        n,k,p,q,no_regimes,TAR_AR_coefs,TAR_MA_coefs,AR_lags,MA_lags,    &
!!        thr_resp_ind,thr_lag,thresholds,TAR_covs,cov_fac_option,extra,x)
!       USE define_types
!       USE stat_utilities
!!     IMPLICIT NONE
!!     INTEGER, INTENT(in) :: max_n, max_k, max_p, max_q, max_no_regimes
!!     INTEGER, INTENT(in) :: n, k, p, q, cov_fac_option, extra
!!     INTEGER, INTENT(in) :: no_regimes, thr_resp_ind, thr_lag
!!     INTEGER, INTENT(in), DIMENSION(p) :: AR_lags
!!     INTEGER, INTENT(in), DIMENSION(q) :: MA_lags
!!     REAL, INTENT(in), DIMENSION(max_no_regimes) :: thresholds
!!     REAL, INTENT(in), DIMENSION(no_regimes, max_k, k) :: TAR_covs
!!     REAL, INTENT(in), DIMENSION(no_regimes, max_p, max_k, k) :: TAR_AR_coefs
!!     REAL, INTENT(in), DIMENSION(no_regimes, max_q, max_k, k) :: TAR_MA_coefs
!!     REAL, INTENT(out), DIMENSION(max_n, k) :: x
!!   END SUBROUTINE gen_mvTAR
!!END INTERFACE
!
  INTERFACE
     SUBROUTINE get_mvSTAR_params(max_k,max_p,max_no_regimes,file_label,    &
          k,p,sm_func,sm_const,sm_param,STAR_AR_coefs,AR_lags,thr_resp_ind, &
          thr_lag)
!       USE define_types
       IMPLICIT NONE
       INTEGER, INTENT(in) :: max_k, max_p, max_no_regimes
       INTEGER, INTENT(in) :: file_label, k
       INTEGER, INTENT(out) :: p, sm_func, thr_resp_ind, thr_lag
       INTEGER, INTENT(out), DIMENSION(max_p) :: AR_lags
       REAL,  INTENT(out) :: sm_param, sm_const
       REAL,  INTENT(out), DIMENSION(max_no_regimes, max_p, max_k, max_k) :: STAR_AR_coefs
     END SUBROUTINE get_mvSTAR_params
  END INTERFACE
!
  INTERFACE
     SUBROUTINE gen_mvSTAR(max_n,max_k,max_p,max_no_regimes,n,k,p,sm_func,   &
          sm_const,sm_param,STAR_AR_coefs,AR_lags,thr_resp_ind,thr_lag,      &
          cov,cov_fac_option,extra,x)
!       USE define_types
!       USE stat_utilities
       IMPLICIT NONE
       INTEGER, INTENT(in) :: max_n, max_k, max_p, max_no_regimes
       INTEGER, INTENT(in) :: n, k, p, extra, cov_fac_option
       INTEGER, INTENT(in) :: sm_func, thr_resp_ind, thr_lag
       INTEGER, INTENT(in), DIMENSION(p) :: AR_lags
       REAL, INTENT(in) :: sm_param, sm_const
       REAL, INTENT(in), DIMENSION(max_k, k) :: cov
       REAL, INTENT(in), DIMENSION(max_no_regimes, max_p, max_k, k) :: STAR_AR_coefs
       REAL, INTENT(out), DIMENSION(max_n, k) :: x
     END SUBROUTINE gen_mvSTAR
  END INTERFACE
!
  INTERFACE
     SUBROUTINE get_mvEXPAR_params(max_k,max_p,file_label,k,p,AR_cons,   &
          AR_coefs,expar_coefs,d)
!       USE define_types
!       USE stat_utilities
       IMPLICIT NONE
       INTEGER, INTENT(in) :: max_k, max_p, file_label, k
       INTEGER, INTENT(out) :: p, d
       REAL, INTENT(out), DIMENSION(max_p, max_k, max_k) :: AR_cons, AR_coefs
       REAL, INTENT(out), DIMENSION(max_p, max_k*max_k, max_k) :: expar_coefs
         END SUBROUTINE get_mvEXPAR_params
  END INTERFACE
!
  INTERFACE
    SUBROUTINE gen_mvEXPAR(max_n,max_k,max_p,n,k,p,AR_cons,AR_coefs,    &
         expar_coefs,d,cov,cov_fac_option,extra,x)
!      USE define_types
!      USE stat_utilities
      IMPLICIT NONE
      INTEGER, INTENT(in) :: max_n, max_k, max_p, n, k, p, d
      INTEGER, INTENT(in) :: cov_fac_option, extra
      REAL, INTENT(in), DIMENSION(max_k, k) :: cov
      REAL, INTENT(in), DIMENSION(max_p, max_k, k) :: AR_cons, AR_coefs
      REAL, INTENT(in), DIMENSION(max_p, max_k*max_k, max_k) :: expar_coefs
      REAL, INTENT(out), DIMENSION(max_n, k) :: x
    END SUBROUTINE gen_mvEXPAR
  END INTERFACE
!
END MODULE nlts_interfaces


MODULE f90_utilities
!**********************************************************************
!   FORTRAN 90 module for defining necessary utilities for various
!   mathematical computations.
!
!   Created: 10/28/2002 JLH
!
!   Notes:  Some subroutines require BLAS subroutines xGEMV, xGEMM,
!           or BLAS function xdot.  All are available for free download
!           from http://www.netlib.org/blas/ .
!           Some subroutines require LAPACK subroutines xGETRF and
!           xGETRI.  Both are available for free download from
!           http://www.netlib.org/lapack/ .
!           In notation above, the lower-case "x" preceding the
!           subroutine or function name is replaced by an "S" for
!           single precision calls or a "D" for double precision
!           calls.
!
!   Contents:  
!         Routines that move data:
!                       swap
!         Argument checking and error handling: 
!			assert_eq, error
!         Routines for matrix opertaions:
!                       mat_mult, vech, cholesky_fac
!
!**********************************************************************
!  USE define_types
  IMPLICIT NONE
!**********************************************************************
!   Routines for moving data:
!**********************************************************************
  INTERFACE swap
     MODULE PROCEDURE swap_i, swap_r, swap_rv, swap_d, swap_dv,       &
          swap_c, swap_cv, swap_cm, swap_z, swap_zv, swap_zm,         &
          masked_swap_rs, masked_swap_rv, masked_swap_rm,             &
          masked_swap_rd, masked_swap_dv, masked_swap_dm               
  END INTERFACE swap
!***********************************************************************
!   Routines for multiplying matrices:
!***********************************************************************
  INTERFACE mat_mult
     MODULE PROCEDURE mat_mult_s, mat_mult_d, vv_mult_s, vv_mult_d,   &
          vm_mult_s, vm_mult_d, mv_mult_s  ! , mv_mult_d
  END INTERFACE mat_mult
!***********************************************************************
!  Routines for computing the half-stacking vector:
!***********************************************************************
!!!  INTERFACE vech
!!!     MODULE PROCEDURE vech_r ! , vech_d
!!!  END INTERFACE vech
!***********************************************************************
!  Routines for computing the Cholesky decomposition:
!***********************************************************************
  INTERFACE cholesky_fac
     MODULE PROCEDURE cholesky_fac_r  ! , cholesky_fac_d
  END INTERFACE cholesky_fac
!**********************************************************************
!  Routines for argument checking and error handling (error is not
!  currently overloaded.
!**********************************************************************
  INTERFACE assert_eq
     MODULE PROCEDURE assert_eq2, assert_eq3, assert_eq4, assert_eqn
  END INTERFACE assert_eq
!***********************************************************************
!  Routines for sorting arrays and carrying indices.
!***********************************************************************
  INTERFACE index_sort
     MODULE PROCEDURE i_sort, r_sort  ! , d_sort
  END INTERFACE index_sort
CONTAINS
!**********************************************************************
!   Swap the contents of a and b.
!**********************************************************************
  SUBROUTINE swap_i(a,b)
!    USE define_types
    IMPLICIT NONE
    INTEGER, INTENT(inout) :: a, b
    INTEGER :: temp
    temp = a
    a = b
    b = temp
    RETURN
  END SUBROUTINE swap_i
!
  SUBROUTINE swap_r(a,b)
!    USE define_types
    IMPLICIT NONE
    REAL, INTENT(inout) :: a, b
    REAL :: temp
    temp = a
    a = b
    b = temp
    RETURN
  END SUBROUTINE swap_r
!
  SUBROUTINE swap_rv(a,b)
!    USE define_types
    IMPLICIT NONE
    REAL, DIMENSION(:), INTENT(inout) :: a, b
    REAL, DIMENSION(SIZE(a)) :: temp
    temp = a
    a = b
    b = temp
    RETURN
  END SUBROUTINE swap_rv
!
  SUBROUTINE swap_d(a,b)
!    USE define_types
    IMPLICIT NONE
    REAL, INTENT(inout) :: a, b
    REAL :: temp
    temp = a
    a = b
    b = temp
    RETURN
  END SUBROUTINE swap_d
!
  SUBROUTINE swap_dv(a,b)
 !   USE define_types
    IMPLICIT NONE
    REAL, DIMENSION(:), INTENT(inout) :: a, b
    REAL, DIMENSION(SIZE(a)) :: temp
    temp = a
    a = b
    b = temp
    RETURN
  END SUBROUTINE swap_dv
!
  SUBROUTINE swap_c(a,b)
!    USE define_types
    IMPLICIT NONE
    COMPLEX, INTENT(inout) :: a, b
    COMPLEX :: temp
    temp = a
    a = b
    b = temp
    RETURN
  END SUBROUTINE swap_c
!
  SUBROUTINE swap_cv(a,b)
!    USE define_types
    IMPLICIT NONE
    COMPLEX, DIMENSION(:), INTENT(inout) :: a, b
    COMPLEX, DIMENSION(SIZE(a)) :: temp
    temp = a
    a = b
    b = temp
    RETURN
  END SUBROUTINE swap_cv
!
  SUBROUTINE swap_cm(a,b)
!    USE define_types
    IMPLICIT NONE
    COMPLEX, DIMENSION(:,:), INTENT(inout) :: a, b
    COMPLEX, DIMENSION(SIZE(a,1), SIZE(a,2)) :: temp
    temp = a
    a = b
    b = temp
    RETURN
  END SUBROUTINE swap_cm
!
  SUBROUTINE swap_z(a,b)
!    USE define_types
    IMPLICIT NONE
    COMPLEX, INTENT(inout) :: a, b
    COMPLEX :: temp
    temp = a
    a = b
    b = temp
    RETURN
  END SUBROUTINE swap_z
!
  SUBROUTINE swap_zv(a,b)
!    USE define_types
    IMPLICIT NONE
    COMPLEX, DIMENSION(:), INTENT(inout) :: a, b
    COMPLEX, DIMENSION(SIZE(a)) :: temp
    temp = a
    a = b
    b = temp
    RETURN
  END SUBROUTINE swap_zv
!
  SUBROUTINE swap_zm(a,b)
!    USE define_types
    IMPLICIT NONE
    COMPLEX, DIMENSION(:,:), INTENT(inout) :: a, b
    COMPLEX, DIMENSION(SIZE(a,1), SIZE(a,2)) :: temp
    temp = a
    a = b
    b = temp
    RETURN
  END SUBROUTINE swap_zm
!
  SUBROUTINE masked_swap_rs(a,b,mask)
!    USE define_types
    IMPLICIT NONE
    REAL, INTENT(inout) :: a, b
    LOGICAL, INTENT(in) :: mask
    REAL :: swp
    IF(mask) THEN
       swp = a
       a = b
       b = swp
    END IF
    RETURN
  END SUBROUTINE masked_swap_rs
!
  SUBROUTINE masked_swap_rv(a,b,mask)
!    USE define_types
    IMPLICIT NONE
    REAL, DIMENSION(:), INTENT(inout) :: a, b
    LOGICAL, DIMENSION(:), INTENT(in) :: mask
    REAL, DIMENSION(SIZE(a)) :: swp
    WHERE(mask)
       swp = a
       a = b
       b = swp
    END WHERE
    RETURN
  END SUBROUTINE masked_swap_rv
!
  SUBROUTINE masked_swap_rm(a,b,mask)
!    USE define_types
    IMPLICIT NONE
    REAL, DIMENSION(:,:), INTENT(inout) :: a, b
    LOGICAL, DIMENSION(:,:), INTENT(in) :: mask
    REAL, DIMENSION(SIZE(a,1),SIZE(a,2)) :: swp
    WHERE(mask)
       swp = a
       a = b
       b = swp
    END WHERE
    RETURN
  END SUBROUTINE masked_swap_rm
!
  SUBROUTINE masked_swap_rd(a,b,mask)
!    USE define_types
    IMPLICIT NONE
    REAL, INTENT(inout) :: a, b
    LOGICAL, INTENT(in) :: mask
    REAL :: swp
    IF(mask) THEN
       swp = a
       a = b
       b = swp
    END IF
    RETURN
  END SUBROUTINE masked_swap_rd
!
  SUBROUTINE masked_swap_dv(a,b,mask)
!    USE define_types
    IMPLICIT NONE
    REAL, DIMENSION(:), INTENT(inout) :: a, b
    LOGICAL, DIMENSION(:), INTENT(in) :: mask
    REAL, DIMENSION(SIZE(a)) :: swp
    WHERE(mask)
       swp = a
       a = b
       b = swp
    END WHERE
    RETURN
  END SUBROUTINE masked_swap_dv
!
  SUBROUTINE masked_swap_dm(a,b,mask)
!    USE define_types
    IMPLICIT NONE
    REAL, DIMENSION(:,:), INTENT(inout) :: a, b
    LOGICAL, DIMENSION(:,:), INTENT(in) :: mask
    REAL, DIMENSION(SIZE(a,1),SIZE(a,2)) ::swp
    WHERE(mask)
       swp = a
       a = b
       b = swp
    END WHERE
    RETURN
  END SUBROUTINE masked_swap_dm
!***********************************************************************
!   Matrix multiplication routines:
!***********************************************************************
  SUBROUTINE mat_mult_s(a,b,c)
!    USE define_types
    IMPLICIT NONE
    REAL, DIMENSION(:,:), INTENT(in) :: a, b
    REAL, DIMENSION(:,:), INTENT(out) :: c
!
    INTEGER :: n, m, k, info
!
    EXTERNAL sgemm
!
    n = SIZE(a,1)
    m = SIZE(a,2)
    k = SIZE(b,2)
!
    info = assert_eq(SIZE(b,1), m,                                     &
         "Matrix dimensions not correct for multiplication")
!
    CALL sgemm("N","N",n,k,m,1.0,a,n,b,m,1.0,c,n)
!
    RETURN
  END SUBROUTINE mat_mult_s
!
  SUBROUTINE mat_mult_d(a,b,c)
!    USE define_types
    IMPLICIT NONE
    REAL, DIMENSION(:,:), INTENT(in) :: a, b
    REAL, DIMENSION(:,:), INTENT(out) :: c
!
    INTEGER :: n, m, k, info
!
    EXTERNAL dgemm
!
    n = SIZE(a,1)
    m = SIZE(a,2)
    k = SIZE(b,2)
!
    info = assert_eq(SIZE(b,1),m,                                     &
         "Matrix dimensions not correct for multiplication.")
!
    CALL dgemm("N","N",n,k,m,1.0,a,n,b,m,1.0,c,n)
!
    RETURN
  END SUBROUTINE mat_mult_d
!
  SUBROUTINE vv_mult_s(a,b,c)
!    USE define_types
    IMPLICIT NONE
    REAL, DIMENSION(:), INTENT(in) :: a, b
    REAL, INTENT(out) :: c
    REAL :: sdot
!
    EXTERNAL sdot
!
    c = sdot(SIZE(a),a,1,b,1)
!
    RETURN
  END SUBROUTINE vv_mult_s
!
  SUBROUTINE vv_mult_d(a,b,c)
!    USE define_types
    IMPLICIT NONE
    REAL, DIMENSION(:), INTENT(in) :: a, b
    REAL, INTENT(out) :: c
    REAL :: ddot
!
    EXTERNAL ddot
!
    c = ddot(SIZE(a),a,1,b,1)
!
    RETURN
  END SUBROUTINE vv_mult_d
!
  SUBROUTINE vm_mult_s(a,b,c)
!    USE define_types
    IMPLICIT NONE
    REAL, DIMENSION(:), INTENT(in) :: a
    REAL, DIMENSION(:,:), INTENT(in) :: b
    REAL, DIMENSION(:,:), INTENT(out) :: c
!
    REAL, DIMENSION(SIZE(a),1) :: work
!
    INTEGER :: info, n, k
!
    EXTERNAL sgemm
!
    n = SIZE(a)
    k = SIZE(b,2)
    info = assert_eq(SIZE(b,1),1,"The matrix must be 1 x k.")
!
    work = RESHAPE(a,(/n, 1/))
!
    CALL sgemm("N","N",n,k,1,1.0,work,n,b,1,1.0,c,n)
!
    RETURN
  END SUBROUTINE vm_mult_s
!
  SUBROUTINE vm_mult_d(a,b,c)
!    USE define_types
    IMPLICIT NONE
    REAL, DIMENSION(:), INTENT(in) :: a
    REAL, DIMENSION(:,:), INTENT(in) :: b
    REAL, DIMENSION(:,:), INTENT(out) :: c
!
    INTEGER :: info, n, k
    REAL, DIMENSION(SIZE(a),1) :: work
!
    EXTERNAL dgemm
!
    n = SIZE(a)
    k = SIZE(b,2)
    info = assert_eq(SIZE(b,1),1,"The matrix must be 1 x k.")
!
    work = RESHAPE(a,(/n, 1/))
!
    CALL dgemm("N","N",n,k,1,1.0,work,n,b,1,1.0,c,n)
!
    RETURN
  END SUBROUTINE vm_mult_d
!
  SUBROUTINE mv_mult_s(a,b,c)
!    USE define_types
    IMPLICIT NONE
    REAL, DIMENSION(:,:), INTENT(in) :: a
    REAL, DIMENSION(:), INTENT(in) :: b
    REAL, DIMENSION(:), INTENT(out) :: c
!
    INTEGER :: info, n, m
!
    EXTERNAL sgemv
!
    n = SIZE(a,1)
    m = SIZE(a,2)
    info = assert_eq(m,SIZE(b),"The matrix must be 1 x m.")
!
    CALL sgemv("N",n,m,1.0,a,n,b,1,1.0,c,1)

    RETURN
  END SUBROUTINE mv_mult_s
!
  SUBROUTINE mv_mult_d(a,b,c)
!    USE define_types
    IMPLICIT NONE
    REAL, DIMENSION(:,:), INTENT(in) :: a
    REAL, DIMENSION(:), INTENT(in) :: b
    REAL, DIMENSION(:), INTENT(out) :: c
!
    INTEGER :: info, n, m
!
    EXTERNAL dgemv
!
    n = SIZE(a,1)
    m = SIZE(a,2)
    info = assert_eq(m,SIZE(b),"The matrix must be 1 x m.")
!
    CALL dgemv("N",n,m,1.0,a,n,b,1,1.0,c,1)
!
    RETURN
  END SUBROUTINE mv_mult_d
!***********************************************************************
!   Half-stacking vector subroutines:
!***********************************************************************
!***********************************************************************
!   FORTRAN subroutine to calculate the stacking or half-stacking
!   vector, v, of the product of two vectors, a and b.
!
!   Input: a = a double precision matrix of dimension n x m.
!          b = a double precision matrix of dimension m x k.
!          option = an integer indicating whether to take the half-
!                   stacking vector (iopt = 0) or the stacking vector
!                   (iopt = 1).
!          n = an integer containing the number of columns of a.
!          k = an integer containing the number of rows of b.
!
!   Output: v = a double precision vector containing the half-stacking
!               or stacking vector of a and b.  If iopt = 0, the 
!               length of y is (nk)[(nk)+1]/2.  If iopt = 1, the 
!               length of z is nk.
!***********************************************************************
  SUBROUTINE vech_r(a,b,n,ndim,m,mdim,k,option,v)
!    USE define_types
    IMPLICIT NONE
    INTEGER, INTENT(in) :: n, ndim, m, mdim, k, option
    REAL, DIMENSION(ndim), INTENT(in) :: a
    REAL, DIMENSION(1,mdim), INTENT(in) :: b
    REAL, DIMENSION(k**2), INTENT(inout) :: v
!  
    INTEGER :: lower_limit, vindex, i, j
!
    v = 0.0
!
    lower_limit = 1
    vindex = 0
!
    DO i = 1, k
       IF(option == 0) lower_limit = i
       DO j = lower_limit, n
          vindex = vindex + 1
          v(vindex) = a(j)*b(1,i)
       END DO
    END DO
    RETURN
  END SUBROUTINE vech_r
!
  SUBROUTINE vech_d(a,b,n,ndim,m,mdim,k,option,v)
!    USE define_types
    IMPLICIT NONE
    INTEGER, INTENT(in) :: n, ndim, m, mdim, k, option
    REAL, DIMENSION(ndim), INTENT(in) :: a
    REAL, DIMENSION(1,mdim), INTENT(in) :: b
    REAL, DIMENSION(k**2), INTENT(inout) :: v
!
    INTEGER :: lower_limit, vindex, i, j
!
    v = 0.0
    
    lower_limit = 1
    vindex = 0
!
    DO i = 1, k
       IF(option == 0) lower_limit = i
       DO j = lower_limit, n
          vindex = vindex + 1
          v(vindex) = a(j)*b(1,i)
       END DO
    END DO
    RETURN
  END SUBROUTINE vech_d
!***********************************************************************
!   Cholesky decomposition subroutines:
!***********************************************************************
!**********************************************************************
!   FORTRAN 90 subroutine for computing the Cholesky factorization of
!   an n x n positive definite matrix a.  The Cholesky factor is 
!   returned in the upper triangular matrix u.
!
!   Created: 10/28/2003 JLH
!**********************************************************************
  SUBROUTINE cholesky_fac_r(x,u)
!    USE define_types
    IMPLICIT NONE
    REAL, DIMENSION(:,:), INTENT(in) :: x
    REAL, DIMENSION(SIZE(x,1),SIZE(x,2)), INTENT(out) :: u
!
    INTEGER :: info, i, j
!
    EXTERNAL spotrf
!
    DO i = 2, SIZE(x,1)
    DO j = 1, i-1
       u(i,j) = 0.0
    END DO
    END DO
!
    CALL spotrf("U",SIZE(x,1),u,SIZE(x,1),info)
!
    RETURN
  END SUBROUTINE cholesky_fac_r
!
  SUBROUTINE cholesky_fac_d(x,u)
!    USE define_types
    IMPLICIT NONE
    REAL, DIMENSION(:,:), INTENT(in) :: x
    REAL, DIMENSION(SIZE(x,1),SIZE(x,2)), INTENT(out) :: u
!
    INTEGER :: info, i, j
!
    EXTERNAL dpotrf
!
    DO i = 2, SIZE(x,1)
    DO j = 1, i-1
       u(i,j) = 0.0
    END DO
    END DO
!
    CALL dpotrf("U",SIZE(x,1),u,SIZE(x,1),info)
!
    RETURN
  END SUBROUTINE cholesky_fac_d
!***********************************************************************
!   Error handling routines:
!**********************************************************************
!   Report and die if integers not all equal (used for size checking).
!**********************************************************************
  FUNCTION assert_eq2(n1,n2,string)
!    USE define_types
    IMPLICIT NONE
    CHARACTER(len=*), INTENT(in) :: string
    INTEGER, INTENT(in) :: n1, n2
    INTEGER :: assert_eq2
!
    IF(n1 == n2) THEN
       assert_eq2 = n1
    ELSE
       WRITE(*,*) "Error: an assert_eq failed with this tag:", string
       STOP " Program terminated by assert_eq2."
    END IF
    RETURN
  END FUNCTION assert_eq2
!
  FUNCTION assert_eq3(n1,n2,n3,string)
!    USE define_types
    IMPLICIT NONE
    CHARACTER(len=*), INTENT(in) :: string
    INTEGER, INTENT(in) :: n1, n2, n3
    INTEGER :: assert_eq3
    IF(n1 == n2 .AND. n2 == n3) THEN
       assert_eq3 = n1
    ELSE
       WRITE(*,*) "Error: an assert_eq failed with this tag:",string
       STOP " Program terminated by assert_eq3."
    END IF
    RETURN
  END FUNCTION assert_eq3
!
  FUNCTION assert_eq4(n1,n2,n3,n4,string)
!    USE define_types
    IMPLICIT NONE
    CHARACTER(len=*), INTENT(in) :: string
    INTEGER, INTENT(in) :: n1, n2, n3, n4
    INTEGER :: assert_eq4
    IF(n1 == n2 .AND. n2 == n3 .AND. n3 == n4) THEN
       assert_eq4 = n1
    ELSE
       WRITE(*,*) "Error: an assert_eq failed with this tag:",string
       STOP " Program terminated by assert_eq4."
    END IF
    RETURN
  END FUNCTION assert_eq4
!
  FUNCTION assert_eqn(nn,string)
!    USE define_types
    IMPLICIT NONE
    CHARACTER(len=*), INTENT(in) :: string
    INTEGER, DIMENSION(:), INTENT(in) :: nn
    INTEGER :: assert_eqn
    IF(ALL(nn(2:) == nn(1))) THEN
       assert_eqn = nn(1)
    ELSE
       WRITE(*,*) "Error: an assert_eq failed with this tag:",string
       STOP " Program terminated by assert_eqn."
    END IF
    RETURN
  END FUNCTION assert_eqn
!
  SUBROUTINE error(string)
!***********************************************************************
!   Report a message, then die.
!***********************************************************************
!    USE define_types
    IMPLICIT NONE
    CHARACTER(len=*), INTENT(in) :: string
    WRITE(*,*) "Error: ", string
    STOP " Program terminated by error."
    RETURN
  END SUBROUTINE error
!
  SUBROUTINE i_sort(n,x_in,indices,x_out)
!    USE define_types
!***********************************************************************
!
!   FORTRAN 90 SUBROUTINE to sort an array x_in, maintaining original
!   subscripts in indices and RETURN the sorted values in the array x_out.
!
!   Input: n - an INTEGER containing the length of x_in,
!          x_in - the DOUBLE PRECISION vector of length n to be sorted,
!          indices - an INTEGER vector of length n containing original
!                subscripts.
!
!   Output: indices - an INTEGER vector containing the subscripts of the
!                 unsorted vector x_in
!           x_out - a DOUBLE PRECISION vector of length n containing the
!               sorted values of the vector x_in.
!
!   Subprograms called: NONE.
!
!***********************************************************************
    IMPLICIT none
    INTEGER, INTENT(in) :: n
    INTEGER, DIMENSION(n), INTENT(inout) :: indices
    INTEGER, DIMENSION(n), INTENT(in) :: x_in
    INTEGER, DIMENSION(n), INTENT(out) :: x_out
    INTEGER :: m, j, i
!
    m = n
    IF(m .LE. 1) RETURN
!
    DO j = 1, m
       x_out(j) = x_in(j)
       indices(j) = j
    END DO
!
80  m = m - 1
    IF(m.EQ.0) GOTO 90
!
    i = 0
70  i = i + 1
    IF(i .GT. m) GOTO 80
    IF(x_out(i) .LT. x_out(i+1)) GOTO 70
!
    CALL swap(x_out(i),x_out(i+1))
    CALL swap(indices(i),indices(i+1))
!
    GOTO 70
!
90  RETURN
  END SUBROUTINE i_sort
!
  SUBROUTINE r_sort(n,x_in,indices,x_out)
!    USE define_types
!***********************************************************************
!
!   FORTRAN 90 SUBROUTINE to sort an array x_in, maintaining original
!   subscripts in indices and RETURN the sorted values in the array x_out.
!
!   Input: n - an INTEGER containing the length of x_in,
!          x_in - the DOUBLE PRECISION vector of length n to be sorted,
!          indices - an INTEGER vector of length n containing original
!                subscripts.
!
!   Output: indices - an INTEGER vector containing the subscripts of the
!                 unsorted vector x_in
!           x_out - a DOUBLE PRECISION vector of length n containing the
!               sorted values of the vector x_in.
!
!   Subprograms called: NONE.
!
!***********************************************************************
    IMPLICIT none
    INTEGER, INTENT(in) :: n
    INTEGER, DIMENSION(n), INTENT(inout) :: indices
    REAL, DIMENSION(n), INTENT(in) :: x_in
    REAL, DIMENSION(n), INTENT(out) :: x_out
    INTEGER :: m, j, i
!
    m = n
    IF(m .LE. 1) RETURN
!
    DO j = 1, m
       x_out(j) = x_in(j)
       indices(j) = j
    END DO
!
80  m = m - 1
    IF(m.EQ.0) GOTO 90
!
    i = 0
70  i = i + 1
    IF(i .GT. m) GOTO 80
    IF(x_out(i) .LT. x_out(i+1)) GOTO 70
!
    CALL swap(x_out(i),x_out(i+1))
    CALL swap(indices(i),indices(i+1))
!
    GOTO 70
!
90  RETURN
  END SUBROUTINE r_sort
!
  SUBROUTINE d_sort(n,x_in,indices,x_out)
!    USE define_types
!***********************************************************************
!
!   FORTRAN 90 SUBROUTINE to sort an array x_in, maintaining original
!   subscripts in indices and RETURN the sorted values in the array x_out.
!
!   Input: n - an INTEGER containing the length of x_in,
!          x_in - the DOUBLE PRECISION vector of length n to be sorted,
!          indices - an INTEGER vector of length n containing original
!                subscripts.
!
!   Output: indices - an INTEGER vector containing the subscripts of the
!                 unsorted vector x_in
!           x_out - a DOUBLE PRECISION vector of length n containing the
!               sorted values of the vector x_in.
!
!   Subprograms called: NONE.
!
!***********************************************************************
    IMPLICIT none
    INTEGER, INTENT(in) :: n
    INTEGER, DIMENSION(n), INTENT(inout) :: indices
    REAL, DIMENSION(n), INTENT(in) :: x_in
    REAL, DIMENSION(n), INTENT(out) :: x_out
    INTEGER :: m, j, i
!
    m = n
    IF(m .LE. 1) RETURN
!
    DO j = 1, m
       x_out(j) = x_in(j)
       indices(j) = j
    END DO
!
80  m = m - 1
    IF(m.EQ.0) GOTO 90
!
    i = 0
70  i = i + 1
    IF(i .GT. m) GOTO 80
    IF(x_out(i) .LT. x_out(i+1)) GOTO 70
!
    CALL swap(x_out(i),x_out(i+1))
    CALL swap(indices(i),indices(i+1))
!
    GOTO 70
!
90  RETURN
  END SUBROUTINE d_sort
!
END MODULE f90_utilities

MODULE f90_interfaces
!***********************************************************************
!  FORTRAN 90 module containing interfaces for f90_utilities.
!***********************************************************************
  INTERFACE
     SUBROUTINE swap_i(a,b)
!       USE define_types
       IMPLICIT NONE
       INTEGER, INTENT(inout) :: a,b
     END SUBROUTINE swap_i
  END INTERFACE
!
  INTERFACE
     SUBROUTINE swap_r(a,b)
!       USE define_types
       IMPLICIT NONE
       REAL, INTENT(inout) :: a,b
     END SUBROUTINE swap_r
  END INTERFACE
!
  INTERFACE
     SUBROUTINE swap_d(a,b)
!       USE define_types
       IMPLICIT NONE
       REAL, INTENT(inout) :: a,b
     END SUBROUTINE swap_d
  END INTERFACE
!
  INTERFACE
     SUBROUTINE swap_dv(a,b)
!       USE define_types
       IMPLICIT NONE
       REAL, DIMENSION(:), INTENT(inout) :: a,b
     END SUBROUTINE swap_dv
  END INTERFACE
!
  INTERFACE
     SUBROUTINE swap_c(a,b)
!       USE define_types
       IMPLICIT NONE
       COMPLEX, INTENT(inout) :: a, b
     END SUBROUTINE swap_c
  END INTERFACE
!
  INTERFACE
     SUBROUTINE swap_cv(a,b)
!       USE define_types
       IMPLICIT NONE
       COMPLEX, DIMENSION(:), INTENT(inout) :: a, b
     END SUBROUTINE swap_cv
  END INTERFACE
!
  INTERFACE
     SUBROUTINE swap_cm(a,b)
!       USE define_types
       IMPLICIT NONE
       COMPLEX, DIMENSION(:,:), INTENT(inout) :: a, b
     END SUBROUTINE swap_cm
  END INTERFACE
!
  INTERFACE
     SUBROUTINE swap_z(a,b)
!       USE define_types
       IMPLICIT NONE
       COMPLEX, INTENT(inout) :: a, b
     END SUBROUTINE swap_z
  END INTERFACE
!
  INTERFACE
     SUBROUTINE swap_zv(a,b)
!       USE define_types
       IMPLICIT NONE
       COMPLEX, DIMENSION(:), INTENT(inout) :: a, b
     END SUBROUTINE swap_zv
  END INTERFACE
!
  INTERFACE
     SUBROUTINE swap_zm(a,b)
!       USE define_types
       IMPLICIT NONE
       COMPLEX, DIMENSION(:,:), INTENT(inout) :: a, b
     END SUBROUTINE swap_zm
  END INTERFACE
!
  INTERFACE
     SUBROUTINE masked_swap_rs(a,b,mask)
!       USE define_types
       IMPLICIT NONE
       REAL, INTENT(inout) :: a, b
       LOGICAL, INTENT(in) :: mask
     END SUBROUTINE masked_swap_rs
  END INTERFACE
!
  INTERFACE
     SUBROUTINE masked_swap_rv(a,b,mask)
!       USE define_types
       IMPLICIT NONE
       REAL, DIMENSION(:), INTENT(inout) :: a, b
       LOGICAL, DIMENSION(:), INTENT(in) :: mask
     END SUBROUTINE masked_swap_rv
  END INTERFACE
!
  INTERFACE
     SUBROUTINE masked_swap_rm(a,b,mask)
!       USE define_types
       IMPLICIT NONE
       REAL, DIMENSION(:,:), INTENT(inout) :: a, b
       LOGICAL, DIMENSION(:,:), INTENT(in) :: mask
     END SUBROUTINE masked_swap_rm
  END INTERFACE
!
  INTERFACE
     SUBROUTINE masked_swap_rd(a,b,mask)
!       USE define_types
       IMPLICIT NONE
       REAL, INTENT(inout) :: a, b
       LOGICAL, INTENT(in) :: mask
     END SUBROUTINE masked_swap_rd
  END INTERFACE
!
  INTERFACE
     SUBROUTINE masked_swap_dv(a,b,mask)
!       USE define_types
       IMPLICIT NONE
       REAL, DIMENSION(:), INTENT(inout) :: a, b
       LOGICAL, DIMENSION(:), INTENT(in) :: mask
     END SUBROUTINE masked_swap_dv
  END INTERFACE
!
  INTERFACE
     SUBROUTINE masked_swap_dm(a,b,mask)
!       USE define_types
       IMPLICIT NONE
       REAL, DIMENSION(:,:), INTENT(inout) :: a, b
       LOGICAL, DIMENSION(:,:), INTENT(in) :: mask
     END SUBROUTINE masked_swap_dm
  END INTERFACE
!
  INTERFACE 
     SUBROUTINE mat_mult_s(a,b,c)
!       USE define_types
       IMPLICIT NONE
       REAL, DIMENSION(:,:), INTENT(in) :: a, b
       REAL, DIMENSION(:,:), INTENT(out) :: c
       EXTERNAL sgemm
     END SUBROUTINE mat_mult_s
  END INTERFACE
!
  INTERFACE 
     SUBROUTINE mat_mult_d(a,b,c)
!       USE define_types
       IMPLICIT NONE
       REAL, DIMENSION(:,:), INTENT(in) :: a, b
       REAL, DIMENSION(:,:), INTENT(out) :: c
       EXTERNAL dgemm
     END SUBROUTINE mat_mult_d
  END INTERFACE
!
  INTERFACE 
     SUBROUTINE vv_mult_s(a,b,c)
!       USE define_types
       IMPLICIT NONE
       REAL, DIMENSION(:), INTENT(in) :: a, b
       REAL, INTENT(out) :: c
       REAL :: sdot
       EXTERNAL sdot
     END SUBROUTINE vv_mult_s
  END INTERFACE
!
  INTERFACE 
     SUBROUTINE vv_mult_d(a,b,c)
!       USE define_types
       IMPLICIT NONE
       REAL, DIMENSION(:), INTENT(in) :: a, b
       REAL, INTENT(out) :: c
       REAL :: ddot
       EXTERNAL ddot
     END SUBROUTINE vv_mult_d
  END INTERFACE
!
  INTERFACE
     SUBROUTINE vm_mult_s(a,b,c)
!       USE define_types
       IMPLICIT NONE
       REAL, DIMENSION(:), INTENT(in) :: a
       REAL, DIMENSION(:,:), INTENT(in) :: b
       REAL, DIMENSION(:,:), INTENT(out) :: c
       EXTERNAL sgemm
     END SUBROUTINE vm_mult_s
  END INTERFACE
!
  INTERFACE
     SUBROUTINE vm_mult_d(a,b,c)
!       USE define_types
       IMPLICIT NONE
       REAL, DIMENSION(:), INTENT(in) :: a
       REAL, DIMENSION(:,:), INTENT(in) :: b
       REAL, DIMENSION(:,:), INTENT(out) :: c
       EXTERNAL dgemm
     END SUBROUTINE vm_mult_d
  END INTERFACE
!
  INTERFACE
     SUBROUTINE mv_mult_s(a,b,c)
!       USE define_types
       IMPLICIT NONE
       REAL, DIMENSION(:,:), INTENT(in) :: a
       REAL, DIMENSION(:), INTENT(in) :: b
       REAL, DIMENSION(:), INTENT(out) :: c
       EXTERNAL sgemv
     END SUBROUTINE mv_mult_s
  END INTERFACE
!
  INTERFACE
     SUBROUTINE mv_mult_d(a,b,c)
!       USE define_types
       IMPLICIT NONE
       REAL, DIMENSION(:,:), INTENT(in) :: a
       REAL, DIMENSION(:), INTENT(in) :: b
       REAL, DIMENSION(:), INTENT(out) :: c
       EXTERNAL dgemv
     END SUBROUTINE mv_mult_d
  END INTERFACE
!
!!INTERFACE
!!!  SUBROUTINE vech_r(a,b,n,ndim,m,mdim,k,option,v)
!       USE define_types
!!!    IMPLICIT NONE
!!!    INTEGER, INTENT(in) :: n, ndim, m, mdim, k, option
!!!    REAL, DIMENSION(ndim), INTENT(in) :: a
!!!    REAL, DIMENSION(1,mdim), INTENT(in) :: b
!!!    REAL, DIMENSION(k**2), INTENT(inout) :: v
!!!  END SUBROUTINE vech_r
!!END INTERFACE
!
  INTERFACE
     SUBROUTINE vech_d(a,b,n,ndim,m,mdim,k,option,v)
!       USE define_types
       IMPLICIT NONE
       INTEGER, INTENT(in) :: n, ndim, m, mdim, k, option
       REAL, DIMENSION(ndim), INTENT(in) :: a
       REAL, DIMENSION(1,mdim), INTENT(in) :: b
       REAL, DIMENSION(k**2), INTENT(inout) :: v
     END SUBROUTINE vech_d
  END INTERFACE
!
  INTERFACE
     SUBROUTINE cholesky_fac_r(x,u)
!       USE define_types
       IMPLICIT NONE
       REAL, DIMENSION(:,:), INTENT(in) :: x
       REAL, DIMENSION(SIZE(x,1),SIZE(x,2)), INTENT(out) :: u
       EXTERNAL spotrf
     END SUBROUTINE cholesky_fac_r
  END INTERFACE
!
  INTERFACE
     SUBROUTINE cholesky_fac_d(x,u)
!       USE define_types
       IMPLICIT NONE
       REAL, DIMENSION(:,:), INTENT(in) :: x
       REAL, DIMENSION(SIZE(x,1),SIZE(x,2)), INTENT(out) :: u
       EXTERNAL dpotrf
     END SUBROUTINE cholesky_fac_d
  END INTERFACE
!
  INTERFACE
     FUNCTION assert_eq2(n1,n2,string)
!       USE define_types
       IMPLICIT NONE
       CHARACTER(len=*), INTENT(in) :: string
       INTEGER, INTENT(in) :: n1, n2
       INTEGER :: assert_eq2
     END FUNCTION assert_eq2
  END INTERFACE
!
  INTERFACE
     FUNCTION assert_eq3(n1,n2,n3,string)
!       USE define_types
       IMPLICIT NONE
       CHARACTER(len=*), INTENT(in) :: string
       INTEGER, INTENT(in) :: n1, n2, n3
       INTEGER :: assert_eq3
     END FUNCTION assert_eq3
  END INTERFACE
!
  INTERFACE
     FUNCTION assert_eq4(n1,n2,n3,n4,string)
!       USE define_types
       IMPLICIT NONE
       CHARACTER(len=*), INTENT(in) :: string
       INTEGER, INTENT(in) :: n1, n2, n3, n4
       INTEGER :: assert_eq4
     END FUNCTION assert_eq4
  END INTERFACE
!
  INTERFACE
     FUNCTION assert_eqn(nn,string)
!       USE define_types
       IMPLICIT NONE
       CHARACTER(len=*), INTENT(in) :: string
       INTEGER, DIMENSION(:), INTENT(in) :: nn
       INTEGER :: assert_eqn
     END FUNCTION assert_eqn
  END INTERFACE
!
  INTERFACE
     SUBROUTINE error(string)
!       USE define_types
       IMPLICIT NONE
       CHARACTER(len=*), INTENT(in) :: string
     END SUBROUTINE error
  END INTERFACE
!
  INTERFACE
     SUBROUTINE i_sort(n,x_in,indices,x_out)
 !      USE define_types
       IMPLICIT none
       INTEGER, INTENT(in) :: n
       INTEGER, INTENT(inout) :: indices
       INTEGER, DIMENSION(n), INTENT(in) :: x_in
       INTEGER, DIMENSION(n), INTENT(out) :: x_out
     END SUBROUTINE i_sort
  END INTERFACE
!
  INTERFACE
     SUBROUTINE r_sort(n,x_in,indices,x_out)
!       USE define_types
       IMPLICIT none
       INTEGER, INTENT(in) :: n
       INTEGER, INTENT(inout) :: indices
       REAL, DIMENSION(n), INTENT(in) :: x_in
       REAL, DIMENSION(n), INTENT(out) :: x_out
     END SUBROUTINE r_sort
  END INTERFACE
!
  INTERFACE
     SUBROUTINE d_sort(n,x_in,indices,x_out)
!       USE define_types
       IMPLICIT none
       INTEGER, INTENT(in) :: n
       INTEGER, INTENT(inout) :: indices
       REAL, DIMENSION(n), INTENT(in) :: x_in
       REAL, DIMENSION(n), INTENT(out) :: x_out
     END SUBROUTINE d_sort
  END INTERFACE
END MODULE f90_interfaces


SUBROUTINE fcar_test(max_n,max_k,n,k,rtrim,res0,res1,rss0,rss1,rss0_det,  &
    rss1_det,test_statistic)
!-----------------------------------------------------------------------
!  FORTRAN90 subroutine to compute lack-of-fit test statistic
!
!  Written: 01/22/04 JLH
!-----------------------------------------------------------------------
!  USE define_types
!
  IMPLICIT NONE
  INTEGER, INTENT(in) :: max_n, max_k, n, k
  REAL, INTENT(in) :: rtrim
  REAL, INTENT(inout) :: rss0_det, rss1_det
  REAL, INTENT(in), DIMENSION(max_n, k) :: res0, res1
  REAL, INTENT(inout) :: test_statistic
  REAL, INTENT(inout), DIMENSION(max_k, k) :: rss0, rss1
!
  INTEGER :: dim, info, upper_lim, lower_lim, nn, index, i, j, t
  REAL :: lambda, perc
  INTEGER, DIMENSION(k) :: pivot
  REAL, DIMENSION(k, k) :: temp0, temp1
  REAL, DIMENSION(n, k) :: ures0, ures1
!
  IF(perc .GT. 0.0) THEN
     perc = (100.0 - DBLE(rtrim))/100.0
     upper_lim = perc*n
     perc = 1.0 - perc
     lower_lim = perc*n + 1
     nn = upper_lim - lower_lim
     DO i = 1, nn
        index = lower_lim + i - 1
        ures0(i, 1:k) = res0(index, 1:k)
        ures1(i, 1:k) = res1(index, 1:k)
     END DO
  ELSE
     nn = n
     lower_lim = 1
     upper_lim = n
     ures0(1:n, 1:k) = res0(1:n, 1:k)
     ures1(1:n, 1:k) = res1(1:n, 1:k)
  END IF
!
  rss0_det = 0.0
  rss1_det = 0.0
  IF(k == 1) THEN
     DO t = 1, nn
        rss0_det = rss0_det + ures0(t,1)*ures0(t,1)
        rss1_det = rss1_det + ures1(t,1)*ures1(t,1)
     END DO
  ELSE
!print *, " shape ures0=", shape(ures0)
!print *, " shape Tr(ures0)=", shape(TRANSPOSE(ures0))
     rss0(1:k,1:k) = MATMUL(TRANSPOSE(ures0(1:nn,1:k)),ures0(1:nn,1:k))
!print *, "MATMUL 1 in fcar_test.f90: OK"
     rss1(1:k,1:k) = MATMUL(TRANSPOSE(ures1(1:nn,1:k)),ures1(1:nn,1:k))
!print *, "MATMUL 2 in fcar_test.f90: OK"
     DO dim = 1, k
        rss0_det = rss0_det + rss0(dim,dim)
        rss1_det = rss1_det + rss1(dim,dim)
     END DO
  END IF
  lambda = SQRT(rss1_det/rss0_det)
  test_statistic = (1.0 - SQRT(lambda))/SQRT(lambda)
! 
  RETURN
END SUBROUTINE fcar_test

SUBROUTINE fit_fcar(max_n,n,k,p,y,x,band_width,u,z,smoother,start,grid,f_dim,f_hat)
!  USE define_types
!  USE stat_utilities
!  USE f90_utilities
!***********************************************************************
!   FOTRAN 90 subroutine to fit a functional coefficient autoregressive
!   model to multivariate data.
!
!   Notes: Uses LAPACK subroutines DGETRF to get LU factorization and 
!          DGETRI to get inverse of matrix.  Both available from free
!          download from http://www.netlib.org/lapack/.
!          Uses BLAS subroutine DGEMM to multiply matrices.  Available
!          for free download from http://www.netlib.org/blas/.
!***********************************************************************
  IMPLICIT NONE
  INTEGER, INTENT(in) :: max_n, n, k, p, f_dim, smoother, grid, start
  REAL, INTENT(in) :: band_width
  REAL, INTENT(in), DIMENSION(max_n) :: u, z
  REAL, INTENT(in), DIMENSION(max_n, k) :: y
  REAL, INTENT(in), DIMENSION(max_n, p) :: x
  REAL, INTENT(inout), DIMENSION(max_n, f_dim, k) :: f_hat
!
  INTEGER :: t, i, j, info, p2, index_of_f, l
  INTEGER, DIMENSION(2*p) :: pivot
  INTEGER, DIMENSION(n) :: u_index
  REAL :: s
  REAL, DIMENSION(n+start+grid+1) :: w0, work
  REAL, DIMENSION((2*p), (2*p)) :: s0, temp
  REAL, DIMENSION((2*p), k) :: s1, s2
  REAL, DIMENSION(n, (2*p)) :: dx, work_space
  REAL, DIMENSION(n, k) :: wy
  REAL, DIMENSION(max_n, f_dim, k) :: temp_f
!
  EXTERNAL dgetrf
  EXTERNAL dgetri
!
  p2 = 2*p
  s0 = 0.0
  s1 = 0.0
  DO i = 1, p2
     s0(i,i) = 1.0
  END DO

  dx(1:n, 1:p) = x(1:n, 1:p)
  DO i = start+1, start+grid
     DO t = 1, n
        IF(smoother == 1) THEN
!!           w0(t) = quartic_sm((u(t) - z(i))/band_width)
        ELSE IF(smoother == 2) THEN
!!           w0(t) = Epanechnikov((u(t) - z(i))/band_width)
        ELSE IF(smoother == 3) THEN
!!           w0(t) = cubic_sm((u(t) - z(i))/band_width)
        END IF
        w0(t) = w0(t) / band_width
        work_space(t,1:p) = w0(t) * dx(t,1:p)
        dx(t,(p+1):p2) = (u(t) - z(i))*x(t,1:p)
        work_space(t,(p+1):p2) = w0(t)*dx(t,(p+1):p2)
     END DO
     temp = MATMUL(TRANSPOSE(dx),work_space)
     s0 = (1.0d0/DBLE(n*n))*s0 + temp
     CALL dgetrf(p2,p2,s0,p2,pivot,info)
     CALL dgetri(p2,s0,p2,pivot,work,n,info)
     DO t = 1, n
        wy(t,1:k) = w0(t)* y(t,1:k)
     END DO
     s1 = MATMUL(TRANSPOSE(dx),wy)
     s2 = MATMUL(s0,s1)
     f_hat(i,1:p,1:k) = s2(1:p,1:k)
  END DO
!
RETURN
END SUBROUTINE fit_fcar

