SUBROUTINE fcar_bandwidth(max_n,n,k,p,f_dim,y,x,u,z,smoother,nh,bw_q,ams,bw_vec)
!***********************************************************************
!  FORTRAN 90 subroutine to compute the band width for FCAR modelling.
!  Notes: Uses BLAS function ddot and subroutine dgemv.  Both are
!         available for free download from http://www.netlib.org/blas/
!***********************************************************************
  USE define_types
!
  IMPLICIT NONE
  INTEGER(i4b), INTENT(in) :: max_n, n, k, p, f_dim, smoother, bw_q, nh
  REAL(dp), INTENT(in), DIMENSION(max_n) :: u, z
  REAL(dp), INTENT(in), DIMENSION(max_n, k) :: y
  REAL(dp), INTENT(in), DIMENSION(max_n, p) :: x
  REAL(dp), INTENT(inout), DIMENSION(nh+1) :: bw_vec, ams
!
  INTEGER(i4b) :: m, n1, i, j, t, jj
  REAL(dp) :: dn, h, sse, maxu, minu, range_of_u, a, step
  REAL(dp), DIMENSION(max_n,f_dim,k) :: f_hat
  REAL(dp), DIMENSION(k) :: err, y_hat
!
  dn = DBLE(n)
  m = INT(0.1_dp*dn)
  ams = 0.0_dp
  maxu = MAXVAL(u(1:n))
  minu = MINVAL(u(1:n))
  range_of_u = maxu - minu
  a = 1.0_dp/(DBLE(n)**0.20_dp)
  step = range_of_u*a*1.5_dp/(DBLE(nh))
  bw_vec(1) = 0.5_dp*range_of_u*a
  DO i = 2, nh+1
     bw_vec(i) = bw_vec(i-1) + step
  END DO
!
  DO i = 1, bw_q
     n1 = n - i*m
     DO j = 1, nh+1
        h = bw_vec(j) * (dn/DBLE(n1))**0.2_dp
        CALL fit_fcar(max_n,n1,k,p,y,x,h,u,z,smoother,n1,m,f_dim,f_hat)
        DO t = n1+1, n1+m
           y_hat = MATMUL(x(t,1:p),f_hat(t,1:p,1:k))
           err(1:k) = y(t,1:k) - y_hat(1:k)
           DO jj = 1, k
              ams(j) = ams(j) + err(jj)*err(jj)
           END DO
        END DO
     END DO
  END DO
!
RETURN
END SUBROUTINE fcar_bandwidth
