PROGRAM fcar_simulation
!
 INCLUDE 'CXML_INCLUDE.F90'     !necessary if using Compaq Visual Fortran
!                                 !includes BLAS libraries in compilation.
!  
  USE define_types
  USE f90_utilities
  USE stat_utilities
  USE ts_utilities
  USE nlts_utilities
  USE fcar_interfaces
!-----------------------------------------------------------------------
!  Declare variables
!-----------------------------------------------------------------------
!  Parameters.  IF larger sizes are needed, make changes here.  They
!  will carry throughout remainder of programs.
!  max_n: Largest allowable sample size ((n + extra) .LE. max_n).
!         Also serves as maximum number of bootstrap replications.
!  max_k: Largest allowable number of components in the vector 
!         time series - NOTE: IF max_k is > 9, change I1 in FORMAT
!         statements 10 and 20 at END of program to be IX where X is
!         the number of digits in max_k (k .LE max_k).
!  max_p: Largest allowable AR order in all models (p .LE max_p).
!  max_q: Largest allowable MA order in all models (q .LE max_q).
!  max_no_regimes:  Largest allowable number of regimes in TAR model
!                   (no_regimes .LE. max_no_regimes).
!  max_ns: Largest allowable number of sample sizes for study in 
!          simulation
!  max_corrs: Largest allowable number of error cross-correlations for
!             study in simulation.
!  bw_q: Number of subsections of data in bandwidth selection.
!  nh : Number of bandwidths to consider in optimal bandwidth selection.
!-----------------------------------------------------------------------
  INTEGER(i4b), PARAMETER :: max_n = 600, max_k = 5
  INTEGER(i4b), PARAMETER :: max_p = 5, max_q = 5
  INTEGER(i4b), PARAMETER :: max_no_regimes = 3
  INTEGER(i4b), PARAMETER :: bw_q = 4, nh = 50
  CHARACTER(30) :: outfile
!-----------------------------------------------------------------------
!  Simulation variables and arrays:
!-----------------------------------------------------------------------
  INTEGER(i4b) :: i, j, l, no_ns, no_corrs, corr_no, n_no, no_bws, bw_no
  INTEGER(i4b) :: no_reps, rep_no
  INTEGER(i4b), ALLOCATABLE, DIMENSION(:) :: ns
  REAL(dp) :: bw, alpha
  REAL(dp), ALLOCATABLE, DIMENSION(:) :: corrs
  REAL(dp), ALLOCATABLE, DIMENSION(:,:) :: bw_ave_mat
  REAL(dp), ALLOCATABLE, DIMENSION(:,:,:,:) :: emp_rr
!-----------------------------------------------------------------------
!  Variables needed for all models:
!-----------------------------------------------------------------------
  INTEGER(i4b) :: seed_option, model_option, extra, plusone
  INTEGER(i4b), DIMENSION(4) :: seed
!  Error mean, standard deviations, and cross-correlation.
  INTEGER(i4b) :: sd_option, mean_option, intercept_option
  REAL(dp), DIMENSION(max_k) :: sds, means
  REAL(dp) :: rho
!  Covariance, AR order, MA order, AR and MA coefficients
  INTEGER(i4b) :: n, k, p, q
  REAL(dp), DIMENSION(max_k, max_k):: cov
  REAL(dp), DIMENSION(max_p, max_k, max_k) :: AR_coefs
  REAL(dp), DIMENSION(max_q, max_k, max_k) :: MA_coefs
!-----------------------------------------------------------------------
!  Variables unique to vector bilinear:
!-----------------------------------------------------------------------
  INTEGER(i4b) :: cp_option, cp_dim
  REAL(dp), DIMENSION(max_k, ((max_p*max_k)**2)) :: cp_coefs
!-----------------------------------------------------------------------
!  Variables unique to vector threshold ARMA:
!-----------------------------------------------------------------------
  INTEGER(i4b) :: no_regimes, thr_resp_ind, thr_lag
  INTEGER(i4b), DIMENSION(max_p) :: AR_lags
  INTEGER(i4b), DIMENSION(max_q) :: MA_lags
  REAL(dp), DIMENSION(max_no_regimes) :: thresholds
  REAL(dp), DIMENSION(max_no_regimes, max_k, max_k) :: TAR_covs
  REAL(dp), DIMENSION(max_no_regimes, max_p, max_k, max_k) :: TAR_AR_coefs
  REAL(dp), DIMENSION(max_no_regimes, max_q, max_k, max_k) :: TAR_MA_coefs
!-----------------------------------------------------------------------
!  Variables unique to vector smoothed threshold ARMA:
!-----------------------------------------------------------------------
  INTEGER(i4b) :: sm_funct
  REAL(dp) :: sm_const, sm_param
!-----------------------------------------------------------------------
!  Variables unique to vector exponential AR:
!-----------------------------------------------------------------------
  INTEGER(i4b) :: expar_delay
  REAL(dp), DIMENSION(max_p, max_k, max_k) :: AR_cons
  REAL(dp), DIMENSION(max_p, max_k*max_k, max_k) :: expar_coefs
!-----------------------------------------------------------------------
!  Vector time series realization matrix:
!-----------------------------------------------------------------------
  INTEGER(i4b) :: t, dim, beta_dim
  REAL(dp), DIMENSION(max_n, max_k) :: x
!-----------------------------------------------------------------------
!  AIC and BIC vectors:
!-----------------------------------------------------------------------
  INTEGER(i4b) :: aic_option, iii
  INTEGER(i4b) :: AR_fit_order, func_resp_ind, func_delay, max_lag, no_obs
  INTEGER(i4b) :: no_cols, beg_sub, end_sub, max_order, bias_option
  INTEGER(i4b), DIMENSION(1) :: ind
  INTEGER(i4b), DIMENSION(max_p) :: orders
  REAL(dp), DIMENSION(max_p) :: aic, bic
!-----------------------------------------------------------------------
!  Functional variable fitting:
!-----------------------------------------------------------------------
  INTEGER(i4b) :: smoother, start, index_of_u
  INTEGER(i4b), DIMENSION(max_n) :: u_index
  REAL(dp) :: opt_bw, trim
  REAL(dp), DIMENSION(max_n) :: func_var, u
  REAL(dp), DIMENSION(2*max_n) :: z
  REAL(dp), DIMENSION(max_n, max_k) :: y, y_s
  REAL(dp), DIMENSION(max_n, max_k*max_p+1) :: xr, xr_s
  REAL(dp), DIMENSION(max_n, max_p*max_k, max_k) :: f_hat, f_hat_s
!-----------------------------------------------------------------------
!  Bootstrap variables:
!-----------------------------------------------------------------------
  INTEGER(i4b) :: no_bs_reps
  REAL(dp) :: rss0_det, rss1_det, T0, p_value
  REAL(dp), DIMENSION(max_n) :: ts_star
  REAL(dp), DIMENSION(max_k, max_k) :: rss0, rss1
  REAL(dp), DIMENSION(max_n, max_k) :: y0_hat, y1_hat
  REAL(dp), DIMENSION(max_n, max_k) :: y0_hat_s, y1_hat_s
  REAL(dp), DIMENSION(max_k*max_p+1, max_k) :: beta_hat
!-----------------------------------------------------------------------
!  Bandwidth selection variables:
!-----------------------------------------------------------------------
  REAL(dp) :: range, step, band_width
  REAL(dp), DIMENSION(nh+1) :: ams, bw_vec
!-----------------------------------------------------------------------
!  END declare variables.
!  Begin user input.
!-----------------------------------------------------------------------
!
  WRITE(*,*) "What is the name of the output file?"
  READ(*,*) outfile
!
  OPEN(100,file=outfile)
  WRITE(100,600)
  WRITE(100,610)
  WRITE(100,600)
!
  WRITE(*,*) "Enter a seed for the random number generator?"
  WRITE(*,*) "   (1 = Yes)"
  READ(*,*) seed_option
  IF(seed_option == 1) THEN
     seed = 0
     WRITE(*,*) "What is the seed?"
     READ(*,*) seed(1)
     WRITE(100,630) seed(1)
     CALL RANDOM_SEED(put = seed)
  ELSE
     WRITE(100,640)
  END IF
  WRITE(100,*)
!
  WRITE(*,*) "How many sample sizes?"
  READ(*,*) no_ns
  WRITE(100,740) no_ns
  WRITE(100,*)
  ALLOCATE(ns(no_ns))
  DO i = 1, no_ns
10   WRITE(*,540) i
     READ(*,*) ns(i)
     IF(ns(i) .GT. max_n) THEN
        WRITE(*,550) max_n
        GO TO 10
     END IF
  END DO
!
  WRITE(*,*) "How many extra (start-up) observations?"
  READ(*,*) extra
!
  WRITE(*,*) "What is the dimension of the time series?"
  READ(*,*) k
  WRITE(100,655) k
  WRITE(100,*)
!
  WRITE(*,*) "What model should be used to generate the data?"
  WRITE(*,*) " (0 => white noise,"
  WRITE(*,*) "  1 => ARMA,"
  WRITE(*,*) "  2 => bilinear,"
  WRITE(*,*) "  3 => threshold ARMA,"
  WRITE(*,*) "  4 => smoothed threshold AR,"
  WRITE(*,*) "  5 => exponential AR.)"
  READ(*,*) model_option
!
  IF(model_option .NE. 3) THEN
     WRITE(*,*) "How many error cross-correlations?"
     READ(*,*) no_corrs
     WRITE(100,745) no_corrs
     WRITE(100,*)
     ALLOCATE(corrs(no_corrs))
     DO i = 1, no_corrs
20      WRITE(*,560) i
        READ(*,*) corrs(i)
        IF(ABS(corrs(i)) .GT. 1.0_dp) THEN
           WRITE(*,570)
           GO TO 20
        END IF
     END DO
     WRITE(*,*) "Should all standard deviations be 1? (1 => Yes)"
     READ(*,*) sd_option
     sds = 1.0_dp
     IF(sd_option .NE. 1) THEN
        DO i = 1,k
           WRITE(*,520) i
           READ(*,*) sds(i)
        END DO
     END IF
  END IF
!
  IF(model_option == 0) THEN
     WRITE(100,620) k
     WRITE(*,*) "Should the mean be zero (1 => yes)?"
     READ(*,*) mean_option
     means = 0.0_dp
     IF(mean_option .NE. 1) THEN
        DO i = 1,k
           WRITE(*,530) i
           READ(*,*) means(i)
        END DO
     END IF
  ELSE IF(model_option == 1) THEN
     AR_coefs = 0.0_dp
     MA_coefs = 0.0_dp
     CALL get_mvARMA_params(max_k,max_p,max_q,100,k,p,q,AR_coefs,MA_coefs)
  ELSE IF(model_option == 2) THEN
     CALL get_mvbilinear_params(max_k,max_p,max_q,100,k,p,q,cp_option,       &
          AR_coefs,MA_coefs,cp_dim,cp_coefs)
  ELSE IF(model_option == 3) THEN
     no_corrs = 1
     CALL get_mvTAR_params(max_k,max_p,max_q,max_no_regimes,100,k,p,q,       &
          no_regimes,TAR_AR_coefs,TAR_MA_coefs,AR_lags,MA_lags,thr_resp_ind, &
          thr_lag,thresholds,TAR_covs)
  ELSE IF(model_option == 4) THEN
     CALL get_mvSTAR_params(max_k,max_p,max_no_regimes,100,k,p,sm_funct,  &
          sm_const,sm_param,TAR_AR_coefs,AR_lags,thr_resp_ind,thr_lag) 
  ELSE IF(model_option == 5) THEN
     CALL get_mvEXPAR_params(max_k,max_p,100,k,p,AR_cons,AR_coefs, &
          expar_coefs,expar_delay)
  END IF
  WRITE(100,*)
!
  WRITE(*,*) "Use AIC or BIC to select AR fit order? (1 => Yes)"
  READ(*,*) aic_option
  IF(aic_option == 1) THEN
     WRITE(*,*) "Use AIC (0) or BIC (1)?"
     READ(*,*) bias_option
     WRITE(*,*) "What is the maximum AR fit order to be allowed?"
     READ(*,*) max_order
     IF(bias_option == 0) WRITE(100,768) 
     IF(bias_option == 1) WRITE(100,769) 
     WRITE(100,*)
  ELSE
     WRITE(*,*) "What is the AR fit order?"
     READ(*,*) AR_fit_order
     WRITE(100,770) AR_fit_order
  END IF
  WRITE(100,*)
!
  WRITE(*,*) "Include an intercept term when fitting the FCAR model? (1 => Yes)"
  READ(*,*) intercept_option
  IF(intercept_option .NE. 1) THEN
     plusone = 0
     WRITE(100,771)
  ELSE IF(intercept_option == 1) THEN
     plusone = 1
     WRITE(100,772)
  END IF
!
25 WRITE(*,*) "What percentage of FCAR estimates to trim for inference?"
  READ(*,*) trim
  IF(trim .LT. 0.0_dp .OR. trim .GT. 20.0_dp) THEN
     WRITE(*,*) "ERROR: The trimming percentage must be between 0 and 20."
     GOTO 25
  END IF
!
  WRITE(*,*) "How many bootstrap replications?"
  READ(*,*) no_bs_reps
  WRITE(100,660) no_bs_reps
  WRITE(100,*)
!
  WRITE(*,*) "How many simulation replications?"
  READ(*,*) no_reps
  WRITE(100,670) no_reps
  WRITE(100,*)
!
  WRITE(*,*) "Enter nominal level of significance"
  READ(*,*) alpha
  WRITE(100,680) alpha
  WRITE(100,690)
  WRITE(100,600)
!-----------------------------------------------------------------------
!   End of user input.
!   Begin simulation:
!-----------------------------------------------------------------------
!   Correlation loop:
!-----------------------------------------------------------------------
  ALLOCATE(bw_ave_mat(MAX(p,q,2), k))
  ALLOCATE(emp_rr(MAX(p,q,2), k, no_ns, no_corrs))
  bw_ave_mat = 0.0_dp
  emp_rr = 0.0_dp
!
  x = 0.0_dp
  DO corr_no = 1, no_corrs
     IF(model_option .NE. 3) rho = corrs(corr_no)
     WRITE(100,720) rho
     IF(model_option .NE. 3) THEN
        DO i = 1,k
           DO j = i,k
              IF(i == j) THEN
                 cov(i,j) = sds(i)*sds(i)
              ELSE
                 cov(i,j) = sds(i)*sds(j)*rho
                 cov(j,i) = cov(i,j)
              END IF
           END DO
        END DO
        WRITE(100,*)
        WRITE(100,730)
        DO i = 1, k
           WRITE(100,*) (cov(i,j), j = 1,k)
        END DO
     END IF
     WRITE(100,600)
!-----------------------------------------------------------------------
!   Sample size loop:
!-----------------------------------------------------------------------
     DO n_no = 1, no_ns
        n = ns(n_no)
        nn = n + extra
        WRITE(100,775) n
        WRITE(100,815)
        DO rep_no = 1, no_reps
!-----------------------------------------------------------------------
!   Generate realization:
!-----------------------------------------------------------------------
           IF(model_option == 0) THEN
              CALL gen_normal(max_n,max_k,n,k,means,cov,1,x)
           ELSE IF(model_option == 1) THEN
              CALL gen_mvARMA(max_n,max_k,max_p,max_q,n,k,p,q,            &
                   AR_coefs,MA_coefs,cov,1,extra,x)
           ELSE IF(model_option == 2) THEN
              CALL gen_mvbilinear(max_n,max_k,max_p,max_q,n,k,p,q,        &
                   AR_coefs,MA_coefs,cp_option,cp_dim,cp_coefs,1,cov,     &
                   extra,x)
           ELSE IF(model_option == 3) THEN
              CALL gen_mvTAR(max_n,max_k,max_p,max_q,max_no_regimes,n,    &
                   k,p,q,no_regimes,TAR_AR_coefs,TAR_MA_coefs,AR_lags,    &
                   MA_lags,thr_resp_ind,thr_lag,thresholds,TAR_covs,1,    &
                   extra,x)
           ELSE IF(model_option == 4) THEN
              CALL gen_mvSTAR(max_n,max_k,max_p,max_no_regimes,n,k,p,     &
                   sm_funct,sm_const,sm_param,TAR_AR_coefs,AR_lags,    &
                   thr_resp_ind,thr_lag,cov,1,extra,x)  
           ELSE IF(model_option == 5) THEN
              CALL gen_mvEXPAR(max_n,max_k,max_p,n,k,p,AR_cons,AR_coefs,  &
                   expar_coefs,expar_delay,cov,1,extra,x)
           END IF
!-----------------------------------------------------------------------
!   Determine AR fit order if requested (aic_option == 1):
!-----------------------------------------------------------------------
           IF(aic_option == 1) THEN
              orders = 0
              aic = 0.0_dp
              bic = 0.0_dp
              CALL mvaic(max_n,n,k,x,max_order,orders,aic,bic)
              IF(bias_option == 1) THEN
                 ind = MINLOC(bic(1:max_order))
                 AR_fit_order = orders(ind(1))
              ELSE
                 ind = MINLOC(aic(1:max_order))
                 AR_fit_order = orders(ind(1))
              END IF
           END IF
!-----------------------------------------------------------------------
!   Functional variable loop (response index and delay):
!-----------------------------------------------------------------------
           IF(intercept_option == 1) xr(1:n,1) = 1.0_dp
           no_cols = k*AR_fit_order + plusone
           DO func_resp_ind = 1, k
              band_width = 0.0_dp
              DO func_delay = 1, MAX(p,q,2)
                 max_lag = MAX(AR_fit_order, func_delay)
                 no_obs = n - max_lag
                 y(1:no_obs,1:k) = x((max_lag+1):n,1:k)                 
                 DO i = 1,AR_fit_order
                    xr(1:no_obs,(k*(i-1)+(1+plusone)):(i*k+plusone)) = &
					    x((max_lag-i+1):(n-i),1:k)
                 END DO
!
                 beg_sub = 1 + max_lag - func_delay
                 end_sub = no_obs + max_lag - func_delay
                 func_var(1:no_obs) = x(beg_sub:end_sub,func_resp_ind)
!-----------------------------------------------------------------------
!   Compute fitted values from linear regression:
!-----------------------------------------------------------------------
                 beta_dim = max_k*max_p+1
                 CALL ls_regression(max_n,no_obs,k,no_cols,xr,y,beta_dim,beta_hat)
                 y0_hat(1:no_obs,1:k) = MATMUL(xr(1:no_obs,1:no_cols),beta_hat(1:no_cols,1:k))
!-----------------------------------------------------------------------
!   Compute bandwidth, fit FCAR model to data, and compute fitted
!   values from FCAR model.
!-----------------------------------------------------------------------
                 CALL fcar_bandwidth(max_n,no_obs,k,no_cols,max_p*max_k,y,xr, &
                      func_var,func_var,3,nh,bw_q,ams,bw_vec)
                 ind = MINLOC(ams)
                 opt_bw = bw_vec(ind(1))
                 CALL fit_fcar(max_n,no_obs,k,no_cols,y,xr,opt_bw,func_var,   &
                      func_var,3,0,no_obs,max_p*max_k,f_hat)
                 y1_hat(1:no_obs,1:k) = 0.0_dp
                 DO j = 1, no_obs
                 DO l = 1, no_cols
                    y1_hat(j,1:k) = y1_hat(j,1:k) + xr(j,l)*f_hat(j,l,1:k)
                 END DO
                 END DO
!-----------------------------------------------------------------------
!   Sort values according to the ordered values of the functional
!   variable.
!-----------------------------------------------------------------------
                 DO i = 1, no_obs
                    u_index(i) = i
                 END DO
                 z(1:no_obs) = u(1:no_obs)
                 CALL index_sort(no_obs,u,u_index,z)
                 DO i = 1, no_obs
                    index_of_u = u_index(i)
                    y_s(i,1:k) = y(index_of_u,1:k)
                    y0_hat_s(i,1:k) = y0_hat(index_of_u,1:k)
                    y1_hat_s(i,1:k) = y1_hat(index_of_u,1:k)
                    xr_s(i,1:no_cols) = xr(index_of_u,1:no_cols)
                 END DO
!
                 y0_hat_s(1:no_obs,1:k) = y_s(1:no_obs,1:k) - y0_hat_s(1:no_obs,1:k)
                 y1_hat_s(1:no_obs,1:k) = y_s(1:no_obs,1:k) - y1_hat_s(1:no_obs,1:k)
                 T0 = 0.0_dp
                 CALL fcar_test(max_n,max_k,no_obs,k,trim,y0_hat_s,y1_hat_s, &
                      rss0,rss1,rss0_det,rss1_det,T0)
!
!  The unsorted y1_hats are replaced with the residuals from the FCAR model
!  for use in bootstrap_test.
!
                 y1_hat(1:no_obs,1:k) = y(1:no_obs,1:k) - y1_hat(1:no_obs,1:k)

                 CALL bootstrap_test(max_n,max_k,no_obs,k,no_cols,trim,y,xr, &
                      func_var,func_var,y0_hat,y1_hat,opt_bw,no_bs_reps,     &
                      rss0_det,rss1_det,T0,ts_star,p_value)
                 WRITE(100,820) rep_no, func_delay, func_resp_ind,            &
                      AR_fit_order, n, rho, T0, p_value
                 IF(p_value .LT. alpha) THEN
                    emp_rr(func_delay,func_resp_ind,n_no,corr_no) =  1.0_dp + &
                         emp_rr(func_delay, func_resp_ind, n_no, corr_no)
                 END IF
              END DO            ! End func_delay loop
           END DO               ! End func_resp_ind loop
        END DO                  ! End no_rep loop
      END DO                    ! End n_no loop
  END DO                        ! End corr_no loop
  WRITE(100,600)
!
  emp_rr = emp_rr/DBLE(no_reps)
!
  WRITE(100,600)
  WRITE(100,830)
  WRITE(100,840)
  WRITE(100,600)
  DO corr_no = 1, no_corrs
     IF(model_option .NE. 3) rho = corrs(corr_no)
     DO n_no = 1, no_ns
        n = ns(n_no)
        DO func_resp_ind = 1, k
           DO func_delay = 1, MAX(p,q,2)
              WRITE(100,850) func_delay, func_resp_ind, n, rho,   &
                   emp_rr(func_delay, func_resp_ind, n_no, corr_no)
           END DO
        END DO
     END DO
  END DO
  WRITE(100,600)
  CLOSE(100)
  DEALLOCATE(ns, bw_ave_mat, emp_rr)
  IF(model_option .NE. 3) DEALLOCATE(corrs)
!-----------------------------------------------------------------------
!   Format statements for variable input (520 - 570)
!   and for results output (600 - 740)
!-----------------------------------------------------------------------
520 FORMAT("Enter standard deviation ",I1,".")
530 FORMAT("Enter mean ",I1,".")
540 FORMAT("Enter sample size ",I1,".")
550 FORMAT("All sample sizes must be less than ",I6,".")
560 FORMAT("Enter error cross-correlation ",I1,".")
570 FORMAT("All error cross-correlations must be between -1 and +1.")
!-----------------------------------------------------------------------
600 FORMAT("******************************************************************")
610 FORMAT(5X,"Generate realization from vector time series")
620 FORMAT("Vector random normal process of dimension ",I1,".")
630 FORMAT("User-specified seed: ",I8,".")
640 FORMAT("Seed not specified.")
650 FORMAT("Number of extra (start-up) observations generated: ",I5)
655 FORMAT("Time series dimension: ",I1,".")
660 FORMAT("Number of bootstrap replications: ", I4,".")
670 FORMAT("Number of simulation replications: ", I5,".")
680 FORMAT("Nominal level of significance: ",F6.4,".")
690 FORMAT("End of user input.")
700 FORMAT("Bootstrap centered residuals.")
710 FORMAT("Bootstrap centered residuals and compute bootstrap sample.")
712 FORMAT("Test statistic = RSS_0/RSS_1 - 1.")
715 FORMAT("Test statistic = likelihood ratio.")
720 FORMAT("Error cross-correlation: ",F8.6)
730 FORMAT("Covariance matrix:")
740 FORMAT("Number of sample sizes: ",I2,".")
745 FORMAT("Number of error cross-correlations: ",I1,".")
750 FORMAT("BIC selected AR fit order: ",I1,".")
760 FORMAT("AIC selected AR fit order: ",I1,".")
768 FORMAT("AIC used to select AR fit order.")
769 FORMAT("BIC used to select AR fit order.")
770 FORMAT("User-specified AR fit order: ",I1,".")
771 FORMAT("Intercept was not included when fitting FCAR model.")
772 FORMAT("Intercept was included when fitting FCAR model.")
775 FORMAT("Sample size: ",I6,".")
780 FORMAT("Functional variable: X_",I1,"(t - ",I1,").")
800 FORMAT("Resp. Ind., Delay, Bandwidth")
810 FORMAT(5X,I2,6X,I2,2X,F10.6)
815 FORMAT("Rep. no., Delay, Resp. ind., AR ord., n, rho, test stat., p-value")
820 FORMAT(I5,2X,I1,3X,I1,3X,I1,3X,I6,2X,F6.4,2X,F12.6,2X,F12.6)
830 FORMAT("Empirical Rejection Rates")
840 FORMAT("Delay, Resp. ind., n,   rho,     Emp. rr.")
850 FORMAT(2X,I1,9X,I1,4X,I6,1X,F6.4,1X,F8.6)
!-----------------------------------------------------------------------
END PROGRAM fcar_simulation

