SUBROUTINE fit_fcar(max_n,n,k,p,y,x,band_width,u,z,smoother,start,grid,f_dim,f_hat)
  USE define_types
  USE stat_utilities
  USE f90_utilities
!***********************************************************************
!   FOTRAN 90 subroutine to fit a functional coefficient autoregressive
!   model to multivariate data.
!
!   Notes: Uses LAPACK subroutines DGETRF to get LU factorization and 
!          DGETRI to get inverse of matrix.  Both available from free
!          download from http://www.netlib.org/lapack/.
!          Uses BLAS subroutine DGEMM to multiply matrices.  Available
!          for free download from http://www.netlib.org/blas/.
!***********************************************************************
  IMPLICIT NONE
  INTEGER(i4b), INTENT(in) :: max_n, n, k, p, f_dim, smoother, grid, start
  REAL(dp), INTENT(in) :: band_width
  REAL(dp), INTENT(in), DIMENSION(max_n) :: u, z
  REAL(dp), INTENT(in), DIMENSION(max_n, k) :: y
  REAL(dp), INTENT(in), DIMENSION(max_n, p) :: x
  REAL(dp), INTENT(inout), DIMENSION(max_n, f_dim, k) :: f_hat
!
  INTEGER(i4b) :: t, i, j, info, p2, index_of_f, l
  INTEGER(i4b), DIMENSION(2*p) :: pivot
  INTEGER(i4b), DIMENSION(n) :: u_index
  REAL(dp) :: s
  REAL(dp), DIMENSION(n+start+grid+1) :: w0, work
  REAL(dp), DIMENSION((2*p), (2*p)) :: s0, temp
  REAL(dp), DIMENSION((2*p), k) :: s1, s2
  REAL(dp), DIMENSION(n, (2*p)) :: dx, work_space
  REAL(dp), DIMENSION(n, k) :: wy
  REAL(dp), DIMENSION(max_n, f_dim, k) :: temp_f
!
  EXTERNAL dgetrf
  EXTERNAL dgetri
!
  p2 = 2*p
  s0 = 0.0_dp
  s1 = 0.0_dp
  DO i = 1, p2
     s0(i,i) = 1.0_dp
  END DO

  dx(1:n, 1:p) = x(1:n, 1:p)
  DO i = start+1, start+grid
     DO t = 1, n
        IF(smoother == 1) THEN
           w0(t) = quartic_sm((u(t) - z(i))/band_width)
        ELSE IF(smoother == 2) THEN
           w0(t) = Epanechnikov((u(t) - z(i))/band_width)
        ELSE IF(smoother == 3) THEN
           w0(t) = cubic_sm((u(t) - z(i))/band_width)
        END IF
        w0(t) = w0(t) / band_width
        work_space(t,1:p) = w0(t) * dx(t,1:p)
        dx(t,(p+1):p2) = (u(t) - z(i))*x(t,1:p)
        work_space(t,(p+1):p2) = w0(t)*dx(t,(p+1):p2)
     END DO
     temp = MATMUL(TRANSPOSE(dx),work_space)
     s0 = (1.0d0/DBLE(n*n))*s0 + temp
     CALL dgetrf(p2,p2,s0,p2,pivot,info)
     CALL dgetri(p2,s0,p2,pivot,work,n,info)
     DO t = 1, n
        wy(t,1:k) = w0(t)* y(t,1:k)
     END DO
     s1 = MATMUL(TRANSPOSE(dx),wy)
     s2 = MATMUL(s0,s1)
     f_hat(i,1:p,1:k) = s2(1:p,1:k)
  END DO
!
RETURN
END SUBROUTINE fit_fcar
