MODULE stat_interfaces
!***********************************************************************
!  FORTRAN 90 module containing interfaces for statistics utilities in
!  stat_utilities.f90.
!
!  Created: 03/01/03 JLH
!***********************************************************************
  INTERFACE
     SUBROUTINE gen_uniform_r(a,b,n,x)
       USE define_types
       IMPLICIT NONE
       INTEGER(i4b), INTENT(in) :: n
       REAL(sp), INTENT(in) :: a, b
       REAL(sp), INTENT(out), DIMENSION(n) :: x
     END SUBROUTINE gen_uniform_r
  END INTERFACE
!
  INTERFACE
     SUBROUTINE gen_uniform_d(a,b,n,x)
       USE define_types
       IMPLICIT NONE
       INTEGER(i4b), INTENT(in) :: n
       REAL(dp), INTENT(in) :: a, b
       REAL(dp), INTENT(out), DIMENSION(n) :: x
     END SUBROUTINE gen_uniform_d
  END INTERFACE
!
  INTERFACE
     SUBROUTINE gen_normal_r(mean,variance,n,x)
       USE define_types
       IMPLICIT NONE
       INTEGER(i4b), INTENT(in) :: n
       REAL(sp), INTENT(in) :: mean, variance
       REAL(sp), INTENT(out), DIMENSION(n) :: x
     END SUBROUTINE gen_normal_r
  END INTERFACE
!
  INTERFACE
     SUBROUTINE gen_normal_d(mean,variance,n,x)
       USE define_types
       IMPLICIT NONE
       INTEGER(i4b), INTENT(in) :: n
       REAL(dp), INTENT(in) :: mean, variance
       REAL(dp), INTENT(out), DIMENSION(n) :: x
     END SUBROUTINE gen_normal_d
  END INTERFACE
!
  INTERFACE
     SUBROUTINE gen_mvnormal_r(max_n,max_k,n,k,mean,cov,cov_fac_option,x)
       USE define_types
       IMPLICIT NONE
       INTEGER(i4b), INTENT(in) :: max_n, max_k, n, k, cov_fac_option
       REAL(sp), INTENT(in), DIMENSION(k) :: mean
       REAL(sp), INTENT(in), DIMENSION(max_k, k) :: cov
       REAL(sp), INTENT(out), DIMENSION(max_n, k) :: x
       EXTERNAL spotrf
     END SUBROUTINE gen_mvnormal_r
  END INTERFACE
!
  INTERFACE
     SUBROUTINE gen_mvnormal_d(max_n,max_k,n,k,mean,cov,cov_fac_option,x)
       USE define_types
       IMPLICIT NONE
       INTEGER(i4b), INTENT(in) :: max_n, max_k, n, k, cov_fac_option
       REAL(dp), INTENT(in), DIMENSION(k) :: mean
       REAL(dp), INTENT(in), DIMENSION(max_k, k) :: cov
       REAL(dp), INTENT(inout), DIMENSION(max_n, k) :: x
       EXTERNAL dpotrf
     END SUBROUTINE gen_mvnormal_d
  END INTERFACE
!
  INTERFACE
     FUNCTION logistic_sm_r(x, sm_constant, sm_PARAMETER)
       USE define_types
       IMPLICIT NONE
       REAL(sp), INTENT(in) :: sm_constant, sm_PARAMETER, x
       REAL(sp) :: logistic_sm_r
     END FUNCTION logistic_sm_r
  END INTERFACE
!
  INTERFACE
     FUNCTION logistic_sm_d(x, sm_constant, sm_PARAMETER)
       USE define_types
       IMPLICIT NONE
       REAL(dp), INTENT(in) :: sm_constant, sm_PARAMETER, x
       REAL(dp) :: logistic_sm_d
     END FUNCTION logistic_sm_d
  END INTERFACE
!
  INTERFACE
     FUNCTION exponential_sm_r(x, sm_constant, sm_PARAMETER)
       USE define_types
       IMPLICIT NONE
       REAL(sp), INTENT(in) :: sm_constant, sm_PARAMETER, x
       REAL(sp) :: exponential_sm_r
     END FUNCTION exponential_sm_r
  END INTERFACE
!
  INTERFACE
     FUNCTION exponential_sm_d(x, sm_constant, sm_PARAMETER)
       USE define_types
       IMPLICIT NONE
       REAL(dp), INTENT(in) :: sm_constant, sm_PARAMETER, x
       REAL(dp) :: exponential_sm_d
     END FUNCTION exponential_sm_d
  END INTERFACE
!
  INTERFACE
     FUNCTION quad_log_sm_r(x, sm_constant, sm_PARAMETER)
       USE define_types
       IMPLICIT NONE
       REAL(sp), INTENT(in) :: sm_constant, sm_PARAMETER, x
       REAL(sp) :: quad_log_sm_r
     END FUNCTION quad_log_sm_r
  END INTERFACE
!
  INTERFACE
     FUNCTION quad_log_sm_d(x, sm_constant, sm_PARAMETER)
       USE define_types
       IMPLICIT NONE
       REAL(dp), INTENT(in) :: sm_constant, sm_PARAMETER, x
       REAL(dp) :: quad_log_sm_d
     END FUNCTION quad_log_sm_d
  END INTERFACE
!
  INTERFACE
     FUNCTION cubic_sm_r(x)
       USE define_types
       IMPLICIT NONE
       REAL(sp), INTENT(in) :: x
       REAL(sp) :: cubic_sm_r
     END FUNCTION cubic_sm_r
  END INTERFACE
!
  INTERFACE
     FUNCTION cubic_sm_d(x)
       USE define_types
       IMPLICIT NONE
       REAL(dp), INTENT(in) :: x
       REAL(dp) :: cubic_sm_d
     END FUNCTION cubic_sm_d
  END INTERFACE
!
  INTERFACE
     FUNCTION Epanechnikov_r(x)
       USE define_types
       IMPLICIT NONE
       REAL(sp), INTENT(in) :: x
       REAL(sp) :: Epanechnikov_r
     END FUNCTION Epanechnikov_r
  END INTERFACE
!
  INTERFACE
     FUNCTION Epanechnikov_d(x)
       USE define_types
       IMPLICIT NONE
       REAL(dp), INTENT(in) :: x
       REAL(dp) :: Epanechnikov_d
     END FUNCTION Epanechnikov_d
  END INTERFACE
!
  INTERFACE
     FUNCTION quartic_sm_r(x)
       USE define_types
       IMPLICIT NONE
       REAL(sp), INTENT(in) :: x
       REAL(sp) :: quartic_sm_r
     END FUNCTION quartic_sm_r
  END INTERFACE
!
  INTERFACE
     FUNCTION quartic_sm_d(x)
       USE define_types
       IMPLICIT NONE
       REAL(dp), INTENT(in) :: x
       REAL(dp) :: quartic_sm_d
     END FUNCTION quartic_sm_d
  END INTERFACE
!
  INTERFACE
     SUBROUTINE ls_regr_r(max_n,n,p,x,y,beta_hat)
       USE define_types
       IMPLICIT NONE
       INTEGER(i4b), INTENT(in) :: max_n, n, p
       REAL(sp), INTENT(in), DIMENSION(n) :: y
       REAL(sp), INTENT(in), DIMENSION(max_n, p) :: x
       REAL(sp), INTENT(out), DIMENSION(p) :: beta_hat
       EXTERNAL sgetrf
       EXTERNAL sgetri
     END SUBROUTINE ls_regr_r
  END INTERFACE
!
  INTERFACE
     SUBROUTINE ls_regr_d(max_n,n,p,x,y,beta_hat)
       USE define_types
       IMPLICIT NONE
       INTEGER(i4b), INTENT(in) :: max_n, n, p
       REAL(dp), INTENT(in), DIMENSION(n) :: y
       REAL(dp), INTENT(in), DIMENSION(max_n, p) :: x
       REAL(dp), INTENT(out), DIMENSION(p) :: beta_hat
       EXTERNAL dgetrf
       EXTERNAL dgetri
     END SUBROUTINE ls_regr_d
  END INTERFACE
!
  INTERFACE
     SUBROUTINE ls_regr_rmv(max_n,n,k,p,x,y,beta_dim,beta_hat)
       USE define_types
       IMPLICIT NONE
       INTEGER(i4b), INTENT(in) :: max_n, n, k, beta_dim, p
       REAL(sp), INTENT(in), DIMENSION(max_n, p) :: x
       REAL(sp), INTENT(in), DIMENSION(max_n, k) :: y
       REAL(sp), INTENT(inout), DIMENSION(beta_dim, k) :: beta_hat
       EXTERNAL sgetrf
       EXTERNAL sgetri
     END SUBROUTINE ls_regr_rmv
  END INTERFACE
!
  INTERFACE
     SUBROUTINE ls_regr_dmv(max_n,n,k,p,x,y,beta_dim,beta_hat)
       USE define_types
       IMPLICIT NONE
       INTEGER(i4b), INTENT(in) :: max_n, n, k, beta_dim, p
       REAL(dp), INTENT(in), DIMENSION(max_n, p) :: x
       REAL(dp), INTENT(in), DIMENSION(max_n, k) :: y
       REAL(dp), INTENT(inout), DIMENSION(beta_dim, k) :: beta_hat
       EXTERNAL dgetrf
       EXTERNAL dgetri
     END SUBROUTINE ls_regr_dmv
  END INTERFACE
!
END MODULE stat_interfaces
